/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMDir.cpp
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : (See AMDir.h for class description)
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

// Project Include Files
#include <AMDir.h>
#include <AMFsys.h>
#include <AMFile.h>
#include <AMirD.h>

extern Boolean globalCheckOnly;  // if TRUE, do not change files, only check them

extern AMirD *amird;     // main object for control and transfers

// Definitions

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::scanEntryInfo
// Description  : Scan, from local filesystem, information about entry
//
// Inputs       : recursive - is this a recursive scan?
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::scanEntryInfo( Boolean recursive )
{
    // Get the path (and setup local vars)
    AString pth, dirpth;
    EntryType ty;
    Boolean lnk;
    time_t lm;
    unsigned long md, sz;
    localPath(dirpth);

    // Get the entry information
    if ( getEntryInfo((char *)dirpth, ty, lnk, lm, md, sz) == FALSE )
    {
        // The file is gone, log and return
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "Dir removed [%s]", 
                 (char *)dirpth );
        return( FALSE );
    }

    // If this has changed type, report it
    if ( ty != ISDIR )
    {
        // Report change, return
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "File changed type [%s]", 
                 (char *)dirpth );
        return( FALSE );
    }

    // Log scan (if necessary)
    LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Dir scanned [%s]", (char *)dirpth );

    // Update the statistics
    SetInfo( lm, md, lnk );

    // 
    // SUBDIRECTORY SCAN

    // Open the directory for reading
    DIR *dir = opendir( (char *)dirpth );
    if ( dir == NULL )
    {
        // I don't know what the heck this is, error out
        LOGMSG2( Log(), AMIRD_ERROR_LEVEL, "Unable to open directory"
                 " [%s], err (%s)", (char *)*this, strerror(errno) );
        return( FALSE );
    }

    // Add a file seperator to the dirpth
    if ( (dirpth.Length() > 0) && (dirpth.At(dirpth.Length()-1) != FS_SEPERATOR) )
        dirpth += FS_SEPERATOR;

    // Now read all the sub entries
    Queue delQueue( subEntries );
    AMEntry *ent;
    struct dirent *de;
    AString newpath;
    while ( (de = readdir(dir)) != NULL )
    {
        // Create a new pathname
        newpath = dirpth;
        newpath += de->d_name;

        // Ignore this and parent directory entries, temp files, 
        // check if this is a symbolic link (and should ignore)
        if ( (strcmp( de->d_name, "." ) != 0) &&
             (strcmp( de->d_name, ".." ) != 0) &&
             (AMEntry::isTempFilename( de->d_name ) == FALSE) &&
             ((AMFsys::FollowSymbolicLinks() == TRUE) ||
              (isSymbolicLink((char *)newpath) == FALSE)) )
        {
            // Check if existing file (scan or create)
            if ( (ent = (AMEntry *)subEntries.find(de->d_name)) != NULL )
            {
                // Scan (if recurisve or file) and unset delete
                if ( (recursive == TRUE) || (ent->entryType() == ISFILE) )
                    ent->scanEntryInfo( recursive );
                delQueue.deleteObj( ent );
            }
            else
            {
                // Scan, create new entry
                EntryType ty = eType((char *)newpath);
                if ( ty == ISFILE )
                    ent = new AMFile( Fsystem(), this, de->d_name);
                else if ( ty == ISDIR )
                    ent = new AMDir( Fsystem(), this, de->d_name );
                else
                {
                    // Log and error out
                    LOGMSG2( Log(), AMIRD_ERROR_LEVEL, 
                             "Unknown type (%d) for entry [%s]", ty, 
                             (char *)newpath );
                    return( FALSE );
                }

                // OK, now log, scan, and insert
                LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "New entry [%s] ...",
                         (char *)newpath );
                ent->scanEntryInfo( TRUE );
                subEntries.insertObj( ent );
            }
        }
    }

    // Close the directory
    closedir( dir );

    // If exported filesystem
    if ( Fsystem()->Exported() == TRUE )
    {
        // Cleanup the deleted files
        while( (ent = (AMEntry *)delQueue.deQueue()) != NULL )
        {
            // Let the filesystem delete file (book-keeping)
            LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "Delete the entry [%s].",
                     (char *)ent->localPath(pth) );
            if ( ent->remove() == TRUE )
                delete subEntries.deleteObj( ent );
        }
    }
    else
        // Ignore deletes
        delQueue.CleanList();

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::remove
// Description  : Remove an entry from the filesystem
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::remove( void )
{
    // Only do if subdirectory empty
    AMEntry *ent;
    while ( (ent = (AMEntry *)subEntries.deQueue()) != NULL )
        ent->remove();

    // OK, log and remove directory
    AString pth;
    LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "Deleting directory [%s].",
             (char *)localPath(pth) );
    return( rmdir((char *)localPath(pth)) );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::doAnnounce
// Description  : Advertise announcement information
//
// Inputs       : info - buffer to place stats
//                annce - queue of announcements
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::doAnnounce( Buffer & info, BufferList *annce )
{
    static Boolean count = 5;
    if (count <= 0)
    {
        return TRUE;
    }
    count--;

    // Do the generic stuff first
    if ( AMEntry::doAnnounce(info,annce) == FALSE )
        return( FALSE );

    // Add header information
    AString pth;
    Buffer *subannce = new Buffer();
    (*subannce) << Fsystem()->Exporter()
                << (AString &)*Fsystem()
                << relativePath(pth);

    // Do the local generic stuff
    if ( AMEntry::doAnnounce(*subannce,annce) == FALSE )
        return( FALSE );

    // Now add the file/subdir details
    (*subannce) << subEntries.listSize();
    AMEntry *ent;
    subEntries.resetIterator();
    while( (ent = (AMEntry *)subEntries.nextIterator()) != NULL )
    {
        // Insert filename, then announcement information
        (*subannce) << (AString &)*ent << (unsigned char)ent->entryType();
        ent->doAnnounce( *subannce, annce );
    }

    // Insert the announcement on list
    if ( annce != NULL )
    {
        annce->push_back( subannce );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::reconcileAnnounce
// Description  : Reconcile an announcement with received information.
//
// Inputs       : info - information about directory
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::reconcileAnnounce( Buffer & info )
{
    // Get the generic information
    if ( AMEntry::reconcileAnnounce(info) == FALSE )
        return( FALSE );

    // Check if need to create the directory
    AString pth;
    if ( (exists((char *)localPath(pth)) == FALSE) &&
         (makePath(*Fsystem(), relativePath(pth), Mode(), Log()) == FALSE) &&
         (setFileInfo() == FALSE) )
    {
        // Log and error out
        LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Unable to create directory [%s]",
                 (char *)localPath(pth) );
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::process an update announcement
// Description  : Process an update announcement
//
// Inputs       : info - buffer to place stats
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::processAnnouncement( Buffer & info )
{
    // Get number of entries, setup
    Queue delQueue( subEntries );
    unsigned short updates;
    AString fnm, pth, str;
    unsigned char et;
    AMEntry *ent;

    // Get the generic information
    if ( AMEntry::reconcileAnnounce(info) == FALSE )
        return( FALSE );

    // Check if need to create the directory
    if ( (exists((char *)localPath(pth)) == FALSE) &&
         (makePath(*Fsystem(), relativePath(pth), Mode(), Log()) == FALSE) &&
         (setFileInfo() == FALSE) )
    {
        // Log and error out
        LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Unable to create directory [%s]",
                 (char *)localPath(pth) );
        return( FALSE );
    }

    // Walk list of updates
    info >> updates;
    for ( int i=0; i<updates; i++ )
    {
        // Check if existing file (scan or create)
        info >> fnm >> et;
        if ( (ent = (AMEntry *)subEntries.find((char *)fnm)) != NULL )
        {
            // If entry type not changed, process (otherwise delete)
            if ( (EntryType)et == ent->entryType() )
            {
                // Scan entry for modification
                ent->reconcileAnnounce(info);
                delQueue.deleteObj( ent );
            }
            else
            {
                // Log change
                LOGMSG1( Log(), AMIRD_REPORT_LEVEL, 
                         "Entry type change [%s] (deleting).", 
                         (char *)ent->localPath(pth) );
                ent->reconcileAnnounce(info);
            }
        }
        else
        {
            // Create new entry
            if ( (EntryType)et == ISFILE )
                ent = new AMFile( Fsystem(), this, (char *)fnm );
            else if ( (EntryType)et == ISDIR )
                ent = new AMDir( Fsystem(), this, (char *)fnm );
            else
            {
                // Log and error out
                LOGMSG2( Log(), AMIRD_ERROR_LEVEL, "Unknown type (%d) for "
                         "entry [%s]", et, (char *)fnm );
                return( FALSE );
            }

            // Log, insert new entry
            ent->reconcileAnnounce(info);
            LOGMSG2( Log(), AMIRD_REPORT_LEVEL, "New entry [%s,%s].",
                 (char *)ent->localPath(pth), (char *)ent->fileInfo(str) );
            subEntries.insertObj( ent );
        }
    }

    // Cleanup the deleted files
    while( (ent = (AMEntry *)delQueue.deQueue()) != NULL )
    {
        // Let the filesystem delete file (book-keeping)
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "Delete entry [%s].",
                 (char *)ent->localPath(pth) );
        // if only checking, then send a message to
        // the server that it needs to be deleted on
        // this client.
        if (globalCheckOnly)
        {
            Buffer *buf = new Buffer();
            AString locPath;
            ent->localPath (locPath);
            char *cn = getenv ("NAME");
            AString clientName = (cn == NULL) ? "<unknown>" : cn;
            *buf << (unsigned short)AMirD::AMIRD_FS_DELETE << clientName << locPath;
            if (!amird->sendInterface(buf))
            {
                LOGMSG1(Log(), AMIRD_ERROR_LEVEL,
                    "Attempt to broadcast DELETE failed for file %s.",
                    (char*)locPath);
            }
        }
        if (ent->remove() == TRUE)
            delete subEntries.deleteObj( ent );
    }

    // Return successfully
    return( TRUE );
}

//
// Static Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMDir::makePath
// Description  : create the directories towards the path
//
// Inputs       : base - base path
//                rel - relative path (from base)
//                mode - permissions
//                log - reporting log
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMDir::makePath( AString & base, AString & rel, 
                         unsigned long mode, AuditLog *log )
{
    // If needed, create parent
    unsigned short pos;
    if ( (pos = rel.lastOccurrence(FS_SEPERATOR)) != EOS )
    {
        // Create parent string, create it
        AString parent = rel.subString(0, pos-1);
        if ( makePath( base, parent, TEMP_PERMS ) == FALSE ) 
            return( FALSE );
    }

    // Add a file seperator to the dirpth
    AString pth = base;
    if ( (pth.Length() > 0) && (pth.At(pth.Length()-1) != FS_SEPERATOR) )
        pth += FS_SEPERATOR;
    pth += rel;

    // Make the directory
    LOGMSG1( log, AMIRD_REPORT_LEVEL, "Making Directory [%s]", (char *)pth );
    if ( mkdir((char *)pth, mode) == -1 )
    {
        // Log error
        LOGMSG1( log, AMIRD_REPORT_LEVEL, 
                 "Unable to create directory [%s]",(char *)pth );
    }

    // Return successfully
    return( TRUE );
}
