/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMEntry.cpp
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : (See AMEntry.h for class description)
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <errno.h>
#include <sys/types.h>
#include <utime.h>

// Project Include Files
#include <AMEntry.h>
#include <AMFsys.h>
#include <AMDir.h>

// Definitions

//
// Static Attributes
AuditLog * AMEntry::entryLog = NULL;

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::setPending
// Description  : Set this file object to pending update (or not).
//
// Inputs       : pend - pending flag
// Outputs      : the log or NULL if not defined

void AMEntry::setPending( Boolean pend )
{
    // On or Off?
    if ( pend == TRUE )
    {
        // Log the pending flag
        LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Setting Pending ON [%s]", 
                 (char *)*this );

        // Turn on pending flag, insert into pending queue
        pendingUpdate = TRUE;
        if ( Fsystem()->PendingUpdates().find(this) == NULL )
            Fsystem()->PendingUpdates().insertObj(this);
    }
    else
    {
        // Log the pending flag
        LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Setting Pending OFF [%s]", 
                 (char *)*this );

        // Turn off pending flag
        pendingUpdate = FALSE;
        Fsystem()->PendingUpdates().deleteObj( this );
    }

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::Log
// Description  : Get log for the events
//
// Inputs       : none
// Outputs      : the log or NULL if not defined

AuditLog * AMEntry::Log( void )
{
    // If FS log exists, use it
    return( ((fsystem == NULL) || (fsystem->Log() == NULL)) ? 
              entryLog : fsystem->Log() );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::doAnnounce
// Description  : Advertise announcement information
//
// Inputs       : info - buffer to place stats
//                annce - queue of announcements
// Outputs      : none

Boolean AMEntry::doAnnounce( Buffer & info, BufferList *annce )
{
    // Add the information about the file
    info << lastModified << mode << isLink;

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::reconcileAnnounce
// Description  : Reconcile an announcement with received information.
//
// Inputs       : info - update information
//                mods - modified files
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMEntry::reconcileAnnounce( Buffer & info )
{
    // OK, read the entry information
    AString pth;
    time_t lm;
    unsigned long md;
    Boolean lnk;
    info >> lm >> md >> lnk;

    // Check for modification
    if ( ( lm != lastModified ) || (md != mode) || (isLink != lnk) )
    {
        // Set file information, update (if exists)
        lastModified = lm;
        mode = md;
        isLink = lnk;
        if ( exists((char *)localPath(pth)) ) setFileInfo();

        // Log changes
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "File info change [%s]", (char *)pth );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::relativePath
// Description  : Get the relative path (from mount/export) point
//
// Inputs       : none
// Outputs      : none

AString & AMEntry::relativePath( AString & out )
{
    // If no parent
    if ( fsystem == NULL )
        out = *this;
    else if ( parent == NULL ) 
        out = "";
    else
    {
        // add local filename
        parent->relativePath( out );
        if ( (out.Length() > 0) && (out.At(out.Length()-1) != FS_SEPERATOR) )
            out += FS_SEPERATOR;
        out += (AString &)*this;
    }

    // Return the string
    return( out );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::setFileInfo
// Description  : Set the file mode and update times
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMEntry::setFileInfo( void )
{
    // Set permissions
    AString pth;
    if ( chmod((char *)localPath(pth), mode) != 0 )
    {
        // Error out
        LOGMSG2( Log(), AMIRD_ERROR_LEVEL, "Unable set mode [%s](%s)", 
                 (char *)*this, strerror(errno) );
        return( FALSE );
    }
    
    // Set the last modification time (keep trying until successfull)
    int ret;
    struct utimbuf t = { lastModified, lastModified };
    while ( (ret = utime( (char *)pth, &t ) != 0) && (errno == EAGAIN) );
    if ( ret != 0 )
    {
        // Error out
        LOGMSG2( Log(), AMIRD_ERROR_LEVEL, "Unable to set modification "
                 "time [%s](%s)", (char *)*this, strerror(errno) );
        return( FALSE );
    }

    // Now propagate changes up towards the root 
    if ( Parent() != NULL ) Parent()->setFileInfo();

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::findEntry
// Description  : Find an entry by its relative path
//
// Inputs       : relpath - relative path 
// Outputs      : pointer tor entry if found, NULL if not

AMEntry * AMEntry::findEntry( AString & relpath )
{
    // If not dir, then bail
    if ( entryType() != ISDIR ) return( NULL );

    // If we are on last step of search
    AMDir * dir = (AMDir *)this;
    unsigned short pos = relpath.firstOccurrence(FS_SEPERATOR);
    if ( pos == EOS ) 
        return( (AMEntry *)dir->Entries().find((char *)relpath) );

    // Get filename in dir
    AString fn = relpath.subString( 0, pos-1 );
    AMEntry *ent;
    if ( (ent = (AMEntry *)dir->Entries().find((char *)fn )) == NULL )
{
LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Unable find [%s]", (char *)fn );
         return( NULL );
}

    // Get remaining path
    AString rpath = relpath.subString( pos+1 );
    return( ent->findEntry( rpath ) );
}

// 
// Static Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMEntry::getEntryInfo
// Description  : get the entry information 
//
// Inputs       : file - file to test
//                eType - entry type
//                isLink - symbolic link flag
//                time_t lm - last modified
//                md - file mode
//                sz - flie size (if applicable)
// Outputs      : TRUE if exists, FALSE otherwise

Boolean AMEntry::getEntryInfo( const char *file, EntryType & eType,
                               Boolean & isLink, time_t & lm, 
                               unsigned long & md, unsigned long & sz )
{
    // First, get filesystem statistics information
    struct stat einfo;
    if ( stat( file, &einfo ) == -1 ) return( FALSE );

    // Get file/dir information
    eType = (S_ISREG(einfo.st_mode)) ? ISFILE : 
            ((S_ISDIR(einfo.st_mode)) ? ISDIR : ISUNKNOWN);
    isLink = S_ISLNK(einfo.st_mode);
    lm = einfo.st_mtime;
    md = einfo.st_mode;
    sz = einfo.st_size;

    // Return successfully
    return( TRUE );
}
