#ifndef AMEntry_INCLUDED
#define AMEntry_INCLUDED

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMEntry.h
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : This is the abstract base class for mirrored files
//                 and directories in AMirD.
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// *** Include Files ***
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

// Project Include Files
#include <AMirDTypes.h>

// *** Definitions *** 

// Forward References 
class AMFsys;
class AMFile;

// Class Definition AMEntry
class AMEntry : public AString {

  public: // Type definitions

    // Entry Types
    typedef enum EntryTypeEnum {
        ISUNKNOWN = 0, // Unknown type
        ISFILE = 1,    // File
        ISDIR = 2      // Directory
    } EntryType;

  public: // Public Interface

    //
    // Constructors and Destructors

    // Attribute Constructor
    AMEntry( AMFsys *fs, AMEntry *par, const char * file )
      : AString( file ),
        fsystem( fs ),
        parent( par ),
        lastModified( 0 ),
        mode( 0 ),
        isLink( 0 ),
        pendingUpdate( FALSE )
    {
    }

    // Destructor
    virtual ~AMEntry( void )
    {
    }

    //
    // Access Methods

    // Return the parent (imported or exported) filesystem.
    AMFsys * Fsystem( void )
    {
        return( fsystem );
    }

    // Return the parent (dir) of this entry
    AMEntry * Parent( void )
    {
        return( parent );
    }

    // Return the time of last modified for this file
    time_t LastModified( void )
    {
        return( lastModified );
    }

    // Return the file permissions mode of this file
    unsigned long Mode( void )
    {
        return( mode );
    }

    // Is this pending and update?
    Boolean & PendingUpdate( void )
    {
        return( pendingUpdate );
    }

    //
    // Abstract Methods

    virtual EntryType entryType( void ) = 0;
      // what type of entry is this? (ISFILE/ISDIR)

    virtual Boolean scanEntryInfo( Boolean recursive ) = 0;
      // Scan, from local filesystem, information about entry

    virtual Boolean remove( void ) = 0;
      // Remove a file from the filesystem

    virtual Boolean xferBegin( void ) = 0;
      // Process the begining of an xfer

    virtual Boolean xferComplete( Boolean suc ) = 0;
      // Process a completed transfer

    // 
    // Member Methods

    void setPending( Boolean pend );
      // Set this file object to pending update (or not).

    AuditLog * Log( void );
      // Get log for the events

    virtual Boolean doAnnounce( Buffer & info, BufferList *annce );
      // Advertise announcement information.

    virtual Boolean reconcileAnnounce( Buffer & info );
      // Reconcile an announcement with received information.

    // Get the local path to this entry
    AString & localPath( AString & out )
    {
        // If no filesystem, just use string
        if ( fsystem == NULL )
            out = *this;
        else
        {
            // Get relative path, mount/export point
            AString rpath;
            out = (AString &)*fsystem;
            if ( (out.Length() > 0) && (out.At(out.Length()-1) != FS_SEPERATOR) )
                out += FS_SEPERATOR;
            out += relativePath(rpath);
        }

        // Return path
        return( out );
    }

    AString & relativePath( AString & out );
      // Get the relative path (from mount/export) point

    Boolean setFileInfo( void );
      // Set the file mode and update times

    AMEntry * findEntry( AString & relpath );
      // Find an entry by its relative path

    // Is this entry a symbolic link?
    Boolean isSymbolicLink( const char * fn )
    {
        struct stat einfo;
        return( (lstat( (char *)fn, &einfo ) != -1) && 
                (S_ISLNK(einfo.st_mode)) );
    }

    // Create a string describing file information
    AString & fileInfo( AString & fi )
    {
        // Print the info, return
        fi.sprintf( "%c %lu md=%lo", (entryType() == ISDIR) ? 'D' : 'F',
                    lastModified, mode );
        return( fi );
    }

    //
    // Static Member Methods

    static Boolean getEntryInfo( const char *file, EntryType & eType, 
                                 Boolean & isLink, time_t & lm, 
                                 unsigned long & md, unsigned long & sz );
      // get the entry information

    // Does this file exist locally?
    static Boolean exists( const char *fle )
    {
        // First, get filesystem statistics information
        struct stat einfo;
        return( stat(fle, &einfo) != -1 );
    }

    // Touch the file
    static Boolean touch( const char *fl )
    {
        int fh = open(fl, O_CREAT|O_TRUNC|O_WRONLY );
        return( (close(fh) == 0) ? TRUE : FALSE );
    }

    // Does this file exist locally?
    static EntryType eType( const char *fle )
    {
        // First, get filesystem statistics information
        struct stat einfo;
        if ( stat(fle, &einfo) == -1 ) return( ISUNKNOWN );
        if ( (S_ISREG(einfo.st_mode)) ) return( ISFILE );
        return( (S_ISDIR(einfo.st_mode)) ? ISDIR : ISUNKNOWN );
    }

    // Get temp filename
    static AString & createTempFilename( AString & f, AString & out )
    {
        // Create name, make sure it does not already exist
        struct stat s; int i = 0;
        do
            out.sprintf( "%s%s.%d", (char *)f, TEMP_POSTFIX, i++ );
        while ( stat((char *)out,&s) >= 0 );

        // Return the temp filename
        return( out );
    }

    // Is this a temporary filename (created by AMirD)?
    static Boolean isTempFilename( const char *fn )
    {
        return( strstr( fn, TEMP_POSTFIX ) != NULL );
    }

    // Set the Entry Log
    static void setEntryLog( AuditLog *log )
    {
        entryLog = log;
    }

  protected: // Protected Interface and Data

    // Set the information associated with this entry
    void SetInfo( time_t lm, unsigned long md, Boolean lnk )
    {
        lastModified = lm;
        mode = md;
        isLink = lnk;
    }

  private: // Private Interface and Data

    // 
    // Class Attributes

    AMFsys * fsystem;
      // This is the parent (imported or exported) filesystem.

    AMEntry * parent;
      // This is the parent (dir) of this entry

    time_t lastModified;
      // This is the time of last modification.

    unsigned long mode;
      // This is the protection mode for the file.

    Boolean isLink;
      // Is this a symbolic link?

    Boolean pendingUpdate;
      // Is this pending and update?

    //
    // Static Attributes

    static AuditLog * entryLog;
      // This is the log for the entry (defaults ot FS log, if defined)

};
// End Class Definition AMEntry

#endif
