/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMFile.cpp
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : (See AMFile.h for class description)
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <unistd.h>

// Project Include Files
#include <AMFile.h>
#include <AMFsys.h>

extern Boolean globalCheckOnly;  // if TRUE, do not change files, only check them

// Definitions

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::scanEntryInfo
// Description  : Scan, from local filesystem, information about entry
//
// Inputs       : recursive - rescursive scan? (IGNORED)
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::scanEntryInfo( Boolean recursive )
{
    // If import and pending, then don't bother scanning
    if ( (Fsystem()->Exported() == FALSE) && (PendingUpdate() == TRUE) )
        return( TRUE );

    // Get the enry information
    AString pth;
    EntryType ty;
    Boolean lnk;
    time_t lm;
    unsigned long md;
    if ( getEntryInfo((char *)localPath(pth), ty, lnk, 
                      lm, md, size) == FALSE )
    {
        // The file is gone, log and return
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "File removed [%s]", 
                 (char *)localPath(pth) );
        lm = md = size = 0;
        return( FALSE );
    }

    // If this has changed type, report it
    if ( ty != ISFILE )
    {
        // Report change, return
        LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "File changed type [%s]", 
                 (char *)localPath(pth) );
        return( FALSE );
    }

    // Get a hash of the local file (if changed)
    if ( (lm != LastModified()) || (fileHash.Length() == 0) ) 
         HashFile();

    // Set the file information
    SetInfo( lm, md, lnk );

    // Log scan (if necessary)
    LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "File scanned [%s]", 
             (char *)localPath(pth) );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::remove
// Description  : Remove an entry from the filesystem
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::remove( void )
{
    // Remove the files from pending, ...
    setPending( FALSE );
    exRequest = TRUE;
    Fsystem()->deleteEntry( this );

    // OK, log and remove directory (if needed)
    AString pth;
    localPath( pth );
    LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "Deleting file [%s].", (char *)pth );

    if (exists((char *)pth) == FALSE)
    {
        return TRUE;
    }

    if (globalCheckOnly)
    {
        return TRUE;
    }

    return unlink((char*)pth) == 0;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::xferBegin
// Description  : Process the begining of an xfer
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::xferBegin( void )
{
    // Clenaup the timer, if any
    if ( upReqTimer != NULL )
    {
        delete upReqTimer;
        upReqTimer = NULL;
    }

    // Move from pending to in progress, return successfully
    Fsystem()->PendingUpdates().deleteObj( this );
    Fsystem()->InProgress().insertObj( this );
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::xferComplete
// Description  : Process a completed transfer
//
// Inputs       : suc - success or failure flag
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::xferComplete( Boolean suc )
{
    // No longer in progress
    Fsystem()->InProgress().deleteObj( this );

    // Set to pending, not pending
    if ( suc == TRUE )
        setPending( FALSE );
    else if ( Fsystem()->PendingUpdates().find( this ) == NULL )
        setPending( TRUE );
    exRequest = TRUE;

    // scan the file
    scanEntryInfo( TRUE );
    HashFile();

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::doAnnounce
// Description  : Advertise announcement information
//
// Inputs       : info - buffer to place stats
//                annce - queue of announcements
// Outputs      : none

Boolean AMFile::doAnnounce( Buffer & info, BufferList *annce )
{
    // Do the generic stuff first
    if ( AMEntry::doAnnounce(info,annce) == FALSE )
        return( FALSE );

    // Is this file pending?
    Boolean pend = (Fsystem()->PendingUpdates().find((char *)*this) != NULL) ?
                   TRUE : FALSE;

    // Add the information about the file
    info << size << fileHash << pend;

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::reconcileAnnounce
// Description  : Reconcile an announcement with received information.
//
// Inputs       : info - information about file
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::reconcileAnnounce( Buffer & info )
{
    // Scan the file, get the generic information
    if ( AMEntry::reconcileAnnounce(info) == FALSE )
        return( FALSE );

    // Get the file data
    AString pth;
    unsigned long sz;
    Buffer inhash;
    Boolean pend;
    info >> sz >> inhash >> pend;

    // If pending, no need for further requests
    exRequest = TRUE; /*(pend == TRUE) ? FALSE : TRUE;*/

    // If does not exist, just set to pending
    if ( exists((char *)localPath(pth)) == FALSE )       
    {
        // If zero size 
        size = sz;
        if ( size == 0 )
        {
            // touch, then set file information
            touch( (char *)pth );
            setFileInfo();
        }
        else
            setPending( TRUE );
    }
    else
    {
        // If pending update, don't bother, check the size/hash of the file
        if ( (PendingUpdate() == FALSE) && 
             ((size != sz) || (inhash != fileHash)) )
        {
            // Log, set pending
            LOGMSG1( Log(), AMIRD_REPORT_LEVEL, "File modified [%s]",
                     (char *)localPath(pth) );
            setPending( TRUE );
    
            // Update inforation
            size = sz;
        }
    }

    // Return successfully
    return( TRUE );
}

//
// Private Interfaces

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFile::HashFile
// Description  : Hash the file
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFile::HashFile( void )
{
    // Hash the file, check return code
    AString pth;
    if ( AMFsys::Hasher().HashFile(localPath(pth), fileHash) == FALSE )
    {
        // Error out
        LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Unable to hash file [%s]", 
                 (char *)*this );
        return( FALSE );
    }

    // Set the last hash, return successfully
    lastHashed = LastModified();
    return( TRUE );
}
