/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMFsys.cpp
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : (See AMFsys.h for class description)
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files

// Project Include Files
#include <AMFsys.h>
#include <AMFile.h>

// Definitions

//
// Static Class Data
HashFunction * AMFsys::hasher = 
HashFunction::HashFunctionFactory( HashFunction::MD5_HASH_FUNCTION );
Boolean AMFsys::followSymbolicLinks = FALSE;
unsigned short AMFsys::minUpdatePeriod = 30;

//
// Constructors and Destructors

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::AMFsys
// Description  : Attribute Constructor.
//
// Inputs       : ex - exporter
//                lpath - local path
//                pol - group policy
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFsys::AMFsys( AString & ex, AString & lpath, AString & pol, AString & lpol, 
                AString & ath, AuditLog *lg )
  : AString( lpath ),
    log( lg ),
    exporter( ex ),
    policy( pol ),
    locPolicy( lpol ),
    pendingUpdates( Queue::FIFO_QUEUE, FALSE ),
    inProgress( Queue::FIFO_QUEUE, FALSE ),
    lastUpdateTimer( NULL ),
    authCertFile( ath ),
    authCert( NULL )
{
    // Initialize the filesystem
    InitFileSystem( TRUE );

    // Return, no return value
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::AMFsys
// Description  : Import Attribute Constructor.
//
// Inputs       : ex - exporter
//                lpath - local path
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFsys::AMFsys( AString & ex, AString & lpath, AString & lpol, AString & ath, 
                AuditLog *lg )
  : AString( lpath ),
    log( lg ),
    exporter( ex ),
    policy( "" ),
    locPolicy( lpol ),
    pendingUpdates( Queue::FIFO_QUEUE, FALSE ),
    inProgress( Queue::FIFO_QUEUE, FALSE ),
    lastUpdateTimer( NULL ),
    authCertFile( ath ),
    authCert( NULL )
{
    // Initialize the filesystem
    InitFileSystem( FALSE );

    // Return, no return value
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::~AMFsys
// Description  : Destructor
//
// Inputs       : none
// Outputs      : none

AMFsys::~AMFsys( void )
{
    // Cleanup entries
    delete entries;

    // Return, no return code
    return;
}

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::processUpdateRequest
// Description  : Process an incoming announcment
//
// Inputs       : ent - entry to delete
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFsys::processUpdateRequest( Buffer * buf )
{
    // Get the file associated with this request
    AMFile *fl;
    if ( (fl = findFileById(buf)) != NULL )
    {
        // Add to pending list, reset req timer (If needed)
        if ( pendingUpdates.find(fl) == NULL )
        {
            // Queue the request
            LOGMSG1( log, AMIRD_DETAIL_LEVEL, "Queueing req %s", (char *)*fl );
            pendingUpdates.insertObj( fl );
        }
        else
        {
            // Log the fact already queued
            LOGMSG1( log, AMIRD_DETAIL_LEVEL, "Pre-queued req %s", (char *)*fl );
        }

        // Log, return processed here
        return( TRUE );
    }

    // Return not processed here
    return( FALSE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::createFileId
// Description  : Create a file Id buffer
//
// Inputs       : fl - file to create request 
// Outputs      : buffer if successful, NULL otherwise

Buffer * AMFsys::createFileId( AMFile *fl )
{
    // Create buffer, stuff contents
    AString pth;
    Buffer * buf = new Buffer();
    (*buf) << exporter << (AString &)*this << fl->relativePath(pth);

    // Return the created buffer
    return( buf );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::findFileById
// Description  : Find the file associated with this ID
//
// Inputs       : fl - file to create request 
// Outputs      : file if successful, NULL otherwise (not for local fs)

AMFile * AMFsys::findFileById( Buffer *buf )
{
    // Get the data out of the message
    unsigned short pos = buf->Position();
    AString ex, path, rpath;
    (*buf) >> ex >> path >> rpath;

    // If meant for this filesystem and found
    AMFile * fl = (( ex != exporter) || (path != *this)) ? 
                  NULL : (AMFile *)findEntry( rpath );

    // If not found, reset position, return result
    if ( fl == NULL ) buf->Position() = pos;
    return( fl );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::deleteEntry
// Description  : Delete a file from all lists
//
// Inputs       : ent - entry to delete
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFsys::deleteEntry( AMEntry *ent )
{
    // NOTE : NEED TO FIGURE OUT A WAY TO RECOVER FROM THIS CASE 
    // PDM: 12/30/00
    ATK_ASSERT( (inProgress.deleteObj( ent ) == NULL),
                "Deleting inprogress file.  Bad things are about to occur.)" );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::allowUpdate
// Description  : checks whether the request to update should be allowed
//
// Inputs       : none
// Outputs      : TRUE if allowed, FALSE otherwise

Boolean AMFsys::allowUpdate( void )
{
    // If no update timer
    Boolean allow = FALSE;
    if ( (lastUpdateTimer == NULL) || (lastUpdateTimer->expired() == TRUE) )
    {
        // If needs be create, reset timer, allow
        if ( lastUpdateTimer == NULL ) 
            lastUpdateTimer = new AtkTimer( minUpdatePeriod*1000, this );
        lastUpdateTimer->reset();
        allow = TRUE;
    }

    // Return allows (or not)
    return( allow );
}

//
// Private Interfaces

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFsys::InitFileSystem
// Description  : initialize the filesystem and data
//
// Inputs       : auth - authorization certificate filename
//                isexp - is exported flag
// Outputs      : none

Boolean AMFsys::InitFileSystem( Boolean isexp )
{
    // Check to see if mount is a directory (it must be)
    if ( AMEntry::eType((char *)*this) == AMEntry::ISDIR )
        entries = new AMDir( this, NULL, (char *)*this );
    else
    {
        // Log error
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Error: Single file filesystem [%s]",
                 (char *)*this );
        entries = NULL;
    }

    // Scan the filesystem
    scanEntInfo();

    // If needed, load the authority certificate
    if ( (authCertFile.Length() > 0) && 
         ((authCert = Certificate::loadCertificate(authCertFile,isexp)) == NULL) )
    {
        // Error out
        LOGMSG2( log, AMIRD_ERROR_LEVEL, "Unable to acquire filesystem "
                 "authority certificate [%s], (FS=%s).", 
                 (char *)authCertFile, (char *)*this );
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}
