#ifndef AMFsys_INCLUDED
#define AMFsys_INCLUDED

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMFsys.h
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : This is the filesystem object used in the AMirD
//                 application.  Exported filesystems use this object
//                 directly, while imported filesystems use AMImpFsys
//                 objects.
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// *** Include Files ***
#include <list>
#include <AuditLog.h>
#include <HashFunction.h>
#include <AtkTimer.h>
#include <Certificate.h>

// Project Include Files
#include <AMDir.h>

// *** Definitions *** 

// Class Definition AMFsys
class AMFsys : public AString {

  public: // Public Interface

    //
    // Constructors and Destructors

    AMFsys( AString & ex, AString & lpath, AString & pol, AString & lpol, 
            AString & ath, AuditLog *lg = NULL );
      // Export Attribute Constructor

    AMFsys( AString & ex, AString & lpath, AString & lpol, AString & ath, 
            AuditLog *lg = NULL );
      // Import Attribute Constructor

    virtual ~AMFsys( void );
      // Destructor

    //
    // Access Methods

    // Return the log associated with this filesystem
    AuditLog * Log( void )
    {
        return( log );
    }

    // Get the local path to this entry
    AString & localPath( AString & out )
    {
        // Copy, return the string
        out = (AString &)*this;
        return( out );
    }

    // Return the exporter of this filesystem
    AString & Exporter( void )
    {
        return( exporter );
    }

    // Return the group policy associated with this filesystem
    AString & Policy( void )
    {
        return( policy );
    }

    // Return the local policy associated with this filesystem
    AString & LocPolicy( void )
    {
        return( locPolicy );
    }

    // Return a reference to the pending queue
    Queue & PendingUpdates( void )
    {
        return( pendingUpdates );
    }

    // Return a reference to the pending queue
    Queue & InProgress( void )
    {
        return( inProgress );
    }

    // Retturn the authroty cert filename
    AString & AuthCertFile( void )
    {
        return( authCertFile );
    }

    // Retturn the authroty cert filename
    Certificate * AuthCert( void )
    {
        return( authCert );
    }

    // 
    // Member Methods

    virtual Boolean processUpdateRequest( Buffer * buf );
      // Process an incoming update request

    virtual Buffer * createFileId( AMFile *fl );
      // Create a file id buffer

    virtual AMFile * findFileById( Buffer *buf );
      // Find the file associated with this ID

    // Is this an exported filesystem
    virtual Boolean Exported( void )
    {
        return( TRUE );
    }

    // Scan a filesystem for changes
    Boolean scanEntInfo( void )
    {
        return( (entries != NULL) ? entries->scanEntryInfo(TRUE) : FALSE );
    }

    // Create a queue of announcements
     Boolean createAnnouncements( BufferList *ans )
    {
        Buffer dmy;
        if ( lastUpdateTimer != NULL ) lastUpdateTimer->reset();
        return( (entries != NULL) ? entries->doAnnounce(dmy,ans) : FALSE );
    }

    // Find an entry by its relative path
    AMEntry * findEntry( AString & relpath )
    {
        if ( relpath == "" ) return( entries );
        return( (entries != NULL) ? entries->findEntry(relpath) : FALSE );
    }

    Boolean deleteEntry( AMEntry *ent );
      // Delete a file from all lists

    Boolean allowUpdate( void );
      // checks whether the request to update should be allowed

    //
    // Static Member Methods

    // Return the hash function object for AMirD
    static HashFunction & Hasher( void )
    {
        return( *hasher );
    }

    // Return a reference to the sym link follow flag.
    static Boolean & FollowSymbolicLinks( void )
    {
        return( followSymbolicLinks );
    }

    // Return a reference to the min update period
    static unsigned short & MinUpdatePeriod( void )
    {
        return( minUpdatePeriod );
    }

    // Get the file information out of an announcement
    static Boolean getFsysFromAnnounce( Buffer & buf, AString & ex, AString & fs )
    {
        unsigned short pos = buf.Position();
        buf >> ex >> fs;
        buf.Position() = pos;
        return( TRUE );
    }

  private: // Private Interface and Data

    //
    // Private Interfaces

    Boolean InitFileSystem( Boolean isexp );
      // initialize the filesystem and data

    // 
    // Class Attributes

    AuditLog *log;
      // This is the log used to report information about the file.

    AString exporter;
      // This is the exporter for the filesystem.

    AString policy;
      // This is the group policy associated with this filesystem

    AString locPolicy;
      // This is the local policy associated with this filesystem

    AMEntry *entries;
      // These are the entries of the filesystem.

    Queue pendingUpdates;
      // This is a list of files pending update.

    Queue inProgress;
      // This is a list of file xfers in progress.

    AtkTimer * lastUpdateTimer;
      // This is the time of last update

    AString authCertFile;
      // This is the filename of the authority certificate

    Certificate * authCert;
      // This is the certificate, if any, associated with the export

    //
    // Static Class Data

    static HashFunction * hasher;
      // This is the object used to do the hashing.

    static Boolean followSymbolicLinks;
      // This flag indicates whether symbolic links should be followed.

    static unsigned short minUpdatePeriod;
       // This is the minimum period of update

};

typedef AMFsys* AMFsysPtr;

class AMFsysList : public list<AMFsysPtr>
{
  public:

    // find the given item and set the iterator. 
    // The item matches if the AString value is the same.
    // Return NULL and set iterator to end() if not found.
    AMFsys *find (AMFsys *fs, iterator *iter)
    {
        AString search = *fs;
        for (*iter = begin(); *iter != end(); (*iter)++)
        {
            AMFsys *f = *(*iter);
            if (search == *f)
            {
                return f;
            }
        }
        return NULL;
    }


    // find the given item.  Return NULL on not found.
    AMFsys *find (AMFsys *fs)
    {
        iterator iter;
        return find (fs, &iter);
    }


    // find the given item by file name.  Return NULL on not found.
    AMFsys *find (AString *fileName)
    {
        iterator iter;
        for (iter = begin(); iter != end(); (iter)++)
        {
            AMFsys *f = *iter;
            if (*fileName == *f)
            {
                return f;
            }
        }
        return NULL;
    }

    // find the given item and erase it if found.  Do
    // not do a delete on the object.  Return it if found,
    // else return NULL.
    AMFsys *deleteItem (AMFsys *fs)
    {
        iterator iter;
        AMFsys *ret;     // return value

        ret = find (fs, &iter);
        if (ret != NULL)
        {
            erase (iter);
        }
        return ret;
    }
};

// End Class Definition AMFsys

#endif
