/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMFxfer.cpp
//
//   Created : 09-Oct-00
//   By : Patrick McDaniel
//
//   Description : (See AMFxfer.h for class description)
//
//   *** Change History ***
//
//   09-Oct-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <Random.h>

// Project Include Files
#include <AMFxfer.h>
#include <AMXferMem.h>
#include <AMFsys.h>
#include <AMirDPolicyImpl.h>
#include <AMirD.h>

extern AMirD *amird;     // main object for control and transfers

// Definitions
//#define DROP_TEST

//
// Static Class Attributes
const char * AMFxfer::statusStrings[] = {
  "OK", "TOOMANYTO", "ARGMISMATCH", "BADSTATE",
  "UNIXERROR", "NOCLIENT", "GROUPERROR", "XFERCOMPL"
};
const char * AMFxfer::pstatusStrings[] = {
  "START", "WAIT_START", "INIT_XFER", "SEND_DATA", 
  "XFER_WAIT", "WAIT_END", "GRP_LEAVE", "ERRORED", "WAIT_CHECK_REQ"
};
const char * AMFxfer::messageStrings[] = {
  "START_ANNOUNCE", "DATA_BLOCK", "DATA_ACK", "XFER_FIN", "XFER_CHECK_REQ", "XFER_CHECK"
};

extern Boolean globalCheckOnly;  // if TRUE, do not change files, only check them

//
// Constructors and Destructors

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::AMFxfer
// Description  : Server Constructor (testing, non filesystem)
//
// Inputs       : trans - transport layer object
//                fl - filename of file to xfer
//                gpol - group policy
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFxfer::AMFxfer( ATransport * trans, AString & fl, AString & gpol, AString & lpol, 
                  AuditLog * lg )
  : group( NULL ),
    log( lg ),
    file( (char *)fl ),
    entry( new AMFile(NULL, NULL, (char *)file) ),
    status( OK ),
    pktSize( DEF_PACKET_SIZE ), 
    winSize( DEF_WINDOW_SIZE ), 
    totalPackets( 0 ), 
    totalBlocks( 0 ),
    blkSize( 0 ),
    dTime( 0 ),
    state( START ),
    startTimer( START_WAIT_TO ),
    ackTimer( ACK_WAIT_TO ),
    finTimer( FIN_WAIT_TO ),
    retries( 0 ),
    blockCounter( 0 ),
    sendMask( NULL ),
    dataBufs( NULL ),
    fhandle( -1 ),
    inProgress( FALSE )
{
    // Initialize the group, get file information
    if ( initGroup( trans, &gpol, &lpol ) == FALSE )
        return;

    // If file does not exist, correct type, good info
    if ( (AMEntry::exists((char *)file) == FALSE) || 
         (entry->eType((char *)file) != AMEntry::ISFILE) ||
         (entry->scanEntryInfo(TRUE) == FALSE) )
    {
        // error out
        LOGMSG1( log, FXFER_ERROR_LEVEL, "Bad Xfer file [%s]", (char *)file );
        changeState( ERRORED );
        changeStatus( UNIXERROR );
    }
    else
    {
        // Divide with pkt_size to find total_n_packets to be sent
        totalPackets = entry->fileSize()/pktSize + 
                       (entry->fileSize()%pktSize > 0);
        totalBlocks  = totalPackets/winSize + 
                       (totalPackets%winSize > 0);
    }

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::AMFxfer
// Description  : Client Constructor (testing, non filesystem)
//
// Inputs       : trans - transport layer object
//                fl - filename of file to xfer
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFxfer::AMFxfer( ATransport * trans, AString & fl, AString & lpol, AuditLog * lg )
  : group( NULL ),
    log( lg ),
    file( (char *)fl ),
    entry( new AMFile( NULL, NULL, (char *)fl) ),
    status( OK ),
    pktSize( DEF_PACKET_SIZE ),
    winSize( DEF_WINDOW_SIZE ), 
    totalPackets( 0 ),
    totalBlocks( 0 ),
    blkSize( 0 ),
    dTime( 0 ),
    state( START ),
    startTimer( START_WAIT_TO*(START_RETRIES+1) ),
    ackTimer( ACK_WAIT_TO ),
    finTimer( FIN_WAIT_TO ),
    retries( 0 ),
    blockCounter( 0 ),
    sendMask( NULL ),
    dataBufs( NULL ),
    fhandle( -1 ),
    inProgress( FALSE )
{
    // Initialize the group 
    if ( initGroup( trans, NULL, &lpol ) == FALSE )
        return;

    // Get local path
    entry->localPath( file );

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::AMFxfer
// Description  : Server Constructor
//
// Inputs       : trans - transport layer object
//                ent - file to xfer
//                gpol - group policy
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFxfer::AMFxfer( ATransport * trans, AMFile *ent, AString & gpol, AString & lpol, 
                  AuditLog * lg )
  : group( NULL ),
    log( lg ),
    file( "" ),
    entry( ent ),
    status( OK ),
    pktSize( DEF_PACKET_SIZE ), 
    winSize( DEF_WINDOW_SIZE ), 
    totalPackets( 0 ), 
    totalBlocks( 0 ),
    blkSize( 0 ),
    dTime( 0 ),
    state( START ),
    startTimer( START_WAIT_TO ),
    ackTimer( ACK_WAIT_TO ),
    finTimer( FIN_WAIT_TO ),
    retries( 0 ),
    blockCounter( 0 ),
    sendMask( NULL ),
    dataBufs( NULL ),
    fhandle( -1 ),
    inProgress( FALSE )
{
    // Initialize the group, get file information
    if ( initGroup( trans, &gpol, &lpol ) == FALSE )
        return;

    // If file does not exist
    entry->localPath(file);
    if ( AMEntry::exists((char *)file) == FALSE )
    {
        // error out
        LOGMSG1( log, MAJOR_ERROR, "Xfer file not found [%s]", (char *)file );
        changeState( ERRORED );
        changeStatus( UNIXERROR );
    }

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::AMFxfer
// Description  : Client Constructor
//
// Inputs       : trans - transport layer object
//                p - directory to place object in
//                f - file to xfer
//                lpol - local policy
//                lg - audit log
// Outputs      : none

AMFxfer::AMFxfer( ATransport * trans, AMFile *ent, AString & lpol, AuditLog * lg )
  : group( NULL ),
    polImpl( NULL ),
    log( lg ),
    file( "" ),
    entry( ent ),
    status( OK ),
    pktSize( DEF_PACKET_SIZE ),
    winSize( DEF_WINDOW_SIZE ), 
    totalPackets( 0 ),
    totalBlocks( 0 ),
    blkSize( 0 ),
    dTime( 0 ),
    state( START ),
    startTimer( START_WAIT_TO*(START_RETRIES+1) ),
    ackTimer( ACK_WAIT_TO ),
    finTimer( FIN_WAIT_TO ),
    retries( 0 ),
    blockCounter( 0 ),
    sendMask( NULL ),
    dataBufs( NULL ),
    fhandle( -1 ),
    inProgress( FALSE )
{
    // Initialize the group 
    if ( initGroup( trans, NULL, &lpol ) == FALSE )
        return;

    // Get local path
    entry->localPath( file );

    LOGMSG1( log, AMIRD_DETAIL_LEVEL,
        "Contstructed transfer object for file %s",
        (char*)file);

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::~AMFxfer
// Description  : Destructor
//
// Inputs       : none
// Outputs      : none

AMFxfer::~AMFxfer( void )
{
    LOGMSG1( log, AMIRD_DETAIL_LEVEL,
        "Deleting transfer object for file %s",
        (char*)file);

    // Remove timers (without deletion)
    waitTimers.CleanList();

    // Cleanup file/data blocks
    if ( fhandle != -1 ) finishFile( status == XFERCOMPL );
    if ( dataBufs != NULL )
    {
        // Cleanup data blocks
        for ( unsigned long i=0; i<winSize; i++ ) delete dataBufs[i];
        delete [] dataBufs;
    }

    // Close the group, cleanup
    if ( (group != NULL) && (group->State() == AGroup::AG_INOPERATION) )
    {
        group->Quit();
        delete group;
    }
    delete polImpl;

    // Cleanup attribute set
    attrs.cleanAttributes();

    // delete the file reference
    AMDir *par = (AMDir*)entry->Parent();
    if (par != NULL)
    {
        if (par->Entries().deleteObj (entry) != NULL)
        {
            LOGMSG1 (log, AMIRD_DETAIL_LEVEL, "Removed %s from parent dir",
                (char*)file);
        }
    }

    // remove from pending list if it is there
    AMFsys *sys = entry->Fsystem();
    if (sys->PendingUpdates().deleteObj (entry) != NULL)
    {
        LOGMSG1 (log, AMIRD_DETAIL_LEVEL, "Removed %s from pending list",
            (char*)file);
    }

    // remove from in-progress list if it is there
    if (sys->InProgress().deleteObj (entry) != NULL)
    {
        LOGMSG1 (log, AMIRD_DETAIL_LEVEL, "Removed %s from in-progress list",
            (char*)file);
    }

    // remove the entry itself
    delete entry;

    // Return, no return code
    return;
}

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::doXfer
// Description  : Execute the file xfer state machine
//
// Inputs       : read selected file descriptors
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::doXfer( fd_set * readfds )
{
    // Repeat for ahwile
    do
    {
        // Receive a message, if any
        Buffer *buf = NULL;
        if ( (group->messagePending() == TRUE) || 
             (group->isReadSet(readfds) == TRUE) )
            recvInterface( &buf );
        else
            // Let group go about its work
            group->doGroup();
    
        // If Server, then do server state machine
        if ( group->IsServer() == TRUE )
            serverStateMachine( buf );
        else 
            clientStateMachine( buf );

        // Check for bad state 
        if ( (status == OK) && ((group->State() == AGroup::AG_SHUTDOWN) || 
                                (group->State() == AGroup::AG_ERRORED)) )
        {
            // Print out exception, error out
            LOGMSG2( log, FXFER_ERROR_LEVEL, "Premature shutdown [%s]  State:%d", 
                     (char *)*entry, group->State() );
            changeState( ERRORED );
            changeStatus( GROUPERROR );
            return( FALSE );
        }
    }
    // While not in a stable state
    while ( (group->messagePending() == TRUE) || (stableState() == FALSE) );

    // Return successfully
    return( TRUE );
}

//
// Private Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::initGroup
// Description  : initialize the xfer group
//
// Inputs       : trans - transport layer object
//                gpol - group policy
//                lpol - local policy
// Outputs      : none

Boolean AMFxfer::initGroup( ATransport * trans, AString * gp, AString * lp )
{
    // Attempt to start control group
    try
    {
        // Setup the file attributes
        AString pth, expr;
        entry->localPath(pth);
        expr = entry->Fsystem()->Exporter();
        fileAttr.setAttribute( new StringVar("file", (char *)pth), TRUE );
        attrs.addAttribute( &fileAttr );
        
        // If needed, add the author ID
        Certificate *a = entry->Fsystem()->AuthCert();
        if ( a != NULL )
        {
            // Setup the authority/id attributes
            StringVar *sv = 
                 new StringVar( "authorid", (char *)a->Subject().CommonName() );
            authIdAttr.setAttribute( sv, TRUE );
            attrs.addAttribute( &authIdAttr );
            sv = new StringVar( "author", (char *)entry->Fsystem()->AuthCertFile() );
            authAttr.setAttribute( sv, TRUE );
            attrs.addAttribute( &authAttr );
        }

        // Create the xfer group
        AString noCli("");
        polImpl = new AMirDPolicyImpl( this, log );
        if ( gp == NULL )
            group = new AGroup( trans, *lp, polImpl, &attrs, log );
        else
        {
            // Get the packet and window size from environment
            AString pvl, wvl;
            if ( (Environment::getStringEnvVar("PSIZE", pvl) == FALSE) ||
                 (Environment::getStringEnvVar("WSIZE", wvl) == FALSE) )
                throw new AtkException( "Bad PSIZE or WSIZE env var" );
            pSize.setAttribute( new StringVar(PSIZE_ATTR,(char*)pvl), TRUE );
            attrs.addAttribute( &pSize );
            wSize.setAttribute( new StringVar(WSIZE_ATTR,(char*)wvl), TRUE );
            attrs.addAttribute( &wSize );

            // Setup (server-side) group object
            group = new AGroup( trans, *gp, *lp, noCli, polImpl, &attrs, log );
        }
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        LOGMSG1( log, FXFER_ERROR_LEVEL, "Exception create xfer group [%s]", 
                 e->Description() );
        changeState( ERRORED );
        changeStatus( GROUPERROR );
        return( FALSE );
    }

    // Connect/establish the group, get policy
    if ( group->Connect( FALSE ) == FALSE ) 
    {
        // error out
        LOGMSG0( log, FXFER_ERROR_LEVEL, "Unable to establish xfer context" );
        changeState( ERRORED );
        changeStatus( GROUPERROR );
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::initXfer
// Description  : initialize the transfer
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::initXfer( void )
{
    // If buffers not created 
    if ( dataBufs == NULL )
    {
        // create data blocks
        dataBufs = new Buffer *[winSize];
        for ( unsigned long i=0; i<winSize; i++ ) 
            dataBufs[i] = new Buffer(pktSize);
    }

    // If file not initialized
    if ( fhandle == -1 )
    {
        // Open file (server read, client write)
        if ( group->IsServer() == TRUE )
        {
            fhandle = open( (char *)file, O_RDONLY );
        }
        else
        {
            // Create a temporary filename for target
            AMEntry::createTempFilename( file, tfile );
            LOGMSG1( log, FXFER_DETAIL_LEVEL, "Temporary filename : [%s]",
                 (char *)tfile );
            fhandle = open( (char *)tfile, O_CREAT|O_WRONLY, TEMP_PERMS );
        }

        // Check the result
        if ( fhandle == -1 )
        {
            // Log, error out
            LOGMSG2( log, FXFER_ERROR_LEVEL, "xfer open failed [%s,%s]",
                     (char *)file, strerror(errno) );
            changeState( ERRORED );
            changeStatus( UNIXERROR );
            return( FALSE );
        }
    }

    // Server side, get server group
    if ( group->IsServer() == TRUE ) 
    {
        // Get all active members of the group
        EntityTableType *et = &group->EntityTable();
#if 1
        {
            EntityTableType::iterator iter;
            AEntity *ent;
            for (iter = et->begin(); iter != et->end(); iter++)
            {
                ent = (*iter).second;

                if ( (*ent != group->LocalID()) && ent->InGroup() )
                {
                    xferQueue.insertObj( new AMXferMem(ent,winSize) );
                }
            }
        }
#else
        while ((e = group->NextEntity(&it)) != NULL)
        {
            if ( (*e != group->LocalID()) && e->InGroup() )
            {
                xferQueue.insertObj( new AMXferMem(e,winSize) );
            }
        }
#endif
    
        // Check someone is in the group
        if ( xferQueue.listSize() <= 0 )
        {
            // Log, error out
            LOGMSG1( log, FXFER_ERROR_LEVEL, "Nobody in transfer group [%s]",
                     (char *)file );
            changeState( ERRORED );
            changeStatus( NOCLIENT );
            return( FALSE );
        }
    }

    // Create the masks, block counter, etc.
    sendMask = new BitVector(winSize);
    blockCounter = retries = 0;

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::nextBlock
// Description  : prepare the next block
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::nextBlock( unsigned long blk )
{
    // Reset the members
    AMXferMem *mem;
    xferQueue.resetIterator();
    while ( (mem = (AMXferMem *)xferQueue.nextIterator()) != NULL )
    {
        // Reset ack mask, insert into acknowledge list
        mem->reset();
        ackQueue.insertObj( mem );
    }

    // Setup send mask, block counter
    blockCounter = blk;
    sendMask->SetVector( FALSE );
    retries = 0;

    // Figure out how many packets are in this block
    blkSize = ((blk+1 >= totalBlocks) && (totalPackets%winSize > 0)) ? 
              totalPackets%winSize : winSize;

    for (unsigned long i=blkSize; i < winSize; i++)
        sendMask->Set(i, TRUE);

    // Read the next block of data
    readNextBlock(); // Read next block

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::cnextBlock
// Description  : client prepare the next block
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::cnextBlock( unsigned long blk )
{
    // Write the old data (if needed)
    if ( (blk>0) && (writeNextBlock() == FALSE) ) return( FALSE );

    // Cleanup data
    for ( unsigned long i=0;i<winSize; i++ ) dataBufs[i]->Reset();

    // Setup send mask, block counter
    blockCounter = blk;
    sendMask->SetVector( FALSE );
    retries = 0;

    // Last block?
    if ( blk+1 >= totalBlocks )
    {
        // Figure out how many packets are in this block, ack nosends
        blkSize = (totalPackets%winSize > 0) ? totalPackets%winSize : winSize;
        for ( unsigned long i=blkSize; i<winSize; i++ ) sendMask->Set( i, TRUE );
    }
    else
        // Full block expected
        blkSize = winSize;

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::readNextBlock
// Description  : Read next block from file
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::readNextBlock( void )
{
    // Read the individual blocks
    for ( unsigned long i=0; i<winSize; i++ ) 
    {
        // If not passed the end of the file
        if ( i < blkSize )
        {
            // read the blocks
            unsigned pkno = (blockCounter*winSize)+i, ps = packetSize(pkno);
            int ret;
            while ( ((ret=read(fhandle, (char *)*dataBufs[i], ps)) != (int)ps) &&
                    (errno == EAGAIN) && (ret == -1))
                printf( "EAGAIN %u %d\n", ps, ret );

            if ( ret != (int)ps )
            {
                // Log, error out
                LOGMSG3( log, FXFER_ERROR_LEVEL, "xfer read failed (%d) [%s,%s]",
                         errno, (char *)file, strerror(errno) );
                // attempt to clean up
                close (fhandle);
                fhandle = -1;

                // End the transmission
                sendFin();
                fxrTimerReset( finTimer );
                changeState( WAIT_END );
                retries = 0;

                // Cleanup ack timer
                waitTimers.removeTimer( &ackTimer );

                // changeState( ERRORED );
                // changeStatus( UNIXERROR );
                return( FALSE );
            }

            // Set the buffer length
            dataBufs[i]->Length() = ps;
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::writeNextBlock
// Description  : Write next block to file
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::writeNextBlock( void )
{
    // Read the individual blocks
    for ( unsigned long i=0; i<blkSize; i++ )
    {
        // Write the blocks
        unsigned long pkno = (blockCounter*winSize)+i, ps = packetSize(pkno);

        // If buffer length wrong, bail
        if ( dataBufs[i]->Length() != ps )
        {
            // Log, error out
            LOGMSG3( log, FXFER_ERROR_LEVEL,
                "xfer write mismatch [%s] Buffer Length:%d  packet size:%d",
                (char *)file, dataBufs[i]->Length(), ps );
            changeState( ERRORED );
            changeStatus( ARGMISMATCH );
            return( FALSE );
        }

        // Write block information
        LOGMSG4( log, FXFER_DETAIL_LEVEL, "Write [%s,%d,%d,%d]",
                 (char *)file, blockCounter, i, ps );

        // write each block
        int ret;
        while ( ((ret=write(fhandle, (char *)*dataBufs[i], ps)) != (int)ps) &&
                (errno == EAGAIN) );
        if ( ret != (int)ps )
        {
            // Log, error out
            LOGMSG2( log, FXFER_ERROR_LEVEL, "xfer write failed [%s,%s]",
                     (char *)file, strerror(errno) );
            changeState( ERRORED );
            changeStatus( UNIXERROR );
            return( FALSE );
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::purgeAckers
// Description  : Purge those people who have not correctly acked
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::purgeAckers( void )
{
    // Walk the list of members who have not successfully acked
    AMXferMem *mem;
    while ( (mem = (AMXferMem *)ackQueue.deQueue()) != NULL )
    {
        // Log ejection from the group,
        LOGMSG1( log, FXFER_DETAIL_LEVEL, "Purging unacker [%s]",
                 (char *)*mem->Entity() );

        // FIXME : PDM 10/13/00
        // NOTE: We should do a proper A2 EJECTION here

        // Remove from group list, cleanup object
        xferQueue.deleteObj( mem ); 
        delete mem;
    }

    // Check someone is in the group
    if ( xferQueue.listSize() <= 0 )
    {
        // Log, error out
        LOGMSG1( log, FXFER_ERROR_LEVEL, "Nobody left in transfer group [%s]",
                 (char *)file );
        changeState( ERRORED );
        changeStatus( NOCLIENT );
        return( FALSE );
    }
    
    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::serverStateMachine
// Description  : execute the server state machine
//
// Inputs       : msg - message received, if any
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::serverStateMachine( Buffer * msg )
{
    // If message received
    unsigned short mt = MAX_MSGS;
    if ( msg != NULL )
    {
        // Get message type, log it
        *msg >> mt;
        LOGMSG2( log, FXFER_DETAIL_LEVEL, "Server recv [%s,%p]", 
                 messageString((MessageTypes)mt), msg );
    }

    if (globalCheckOnly)
    {
        switch( state )
        {
            case START: // Initial state
                if ( group->State() == AGroup::AG_INOPERATION )
                {
                    LOGMSG1( log, FXFER_DETAIL_LEVEL,
                        "Server START state machine for file %s",
                        (char*)(*entry));
                        
                    fxrTimerReset(startTimer);
                    changeState (WAIT_START);
                }
            break;

            case WAIT_START:
                if (startTimer.expired())
                {
                    LOGMSG1( log, FXFER_DETAIL_LEVEL,
                        "Server WAIT_START state machine for file %s",
                        (char*)(*entry));
                    changeState (WAIT_CHECK_REQ);
                    fxrTimerReset (ackTimer);

                    Buffer *buf = new Buffer;
                    *buf << (unsigned short)XFER_CHECK_REQ;
                    sendInterface(buf);
                }
            break;

            case WAIT_CHECK_REQ:  // expecting a file check
                if (mt == XFER_CHECK)
                {
                    Buffer fileHash;
                    AString clientName;
                    AString *exportName = entry->Fsystem();
                    *msg >> clientName >> fileHash;
                    AString fileHashText = "";
                    fileHashText.Bin2Ascii ((char*)fileHash.Buf(), fileHash.Length());
                    LOGMSG4 (log, FXFER_REPORT,
                        "export:%s    client:%s    file name:%s    %s",
                        (char*)exportName, (char*)clientName,
                        (char*)file, (char*)fileHashText);

                    if (fileHashText == "00")
                    {
                        // if the hash is invalid, then the file needs to be created
                        LOGMSG2 (amird->Report(), AMIRD_REPORT_LEVEL,
                            "%s create %s", (char*)clientName, (char*)file);
                    }
                    else
                    {
                        LOGMSG3 (amird->Report(), AMIRD_REPORT_LEVEL,
                            "%s replace %s %s", (char*)clientName, 
                            (char*)file, (char*)fileHashText);
                    }

                    // keep waiting as long as we are getting input
                    fxrTimerReset (ackTimer);
                }
                if (ackTimer.expired())
                {
                    fxrTimerReset (finTimer);
                    changeState (WAIT_END);
                }
            break;

            case WAIT_END:
                if (finTimer.expired() || (group->groupSize() <= 1))
                {
                    LOGMSG1( log, FXFER_REPORT_LEVEL, "Completed check [%s]",
                       (char *)file );
                    // Quit the group (non-blocking)
                    group->Quit(FALSE);
                    changeState(GRP_LEAVE);
                }
            break;

            case GRP_LEAVE: // Wait to leave group
                if ( group->State() != AGroup::AG_INOPERATION )
                {
                    LOGMSG1( log, FXFER_DETAIL_LEVEL,
                        "Exited from group [%s]", (char*)file);
                    changeStatus(XFERCOMPL);
                }

            default:
                LOGMSG1( log, FXFER_ERROR_LEVEL,
                    "Bad state in check server [%d]", state );
                changeStatus( BADSTATE );
                changeState( ERRORED );
                break;
        }
        return (status == OK) || (status == XFERCOMPL);
    }

    // Switch on server state
    switch( state )
    {
        case START: // Initial state
             if ( group->State() == AGroup::AG_INOPERATION )
             {
                 // Get policy
                 IntVar pkVar( FX_PKT_SIZE ), wVar( FX_WIN_SIZE );
                 if ( (AMirD::getAttributeParm(group, pkVar, 
                                               FX_PKT_SIZE, log) == FALSE) ||
                       (AMirD::getAttributeParm(group, wVar, 
                                                FX_WIN_SIZE, log) == FALSE) )
                 {
                     // Error on bad policy
                     LOGMSG1( log, FXFER_ERROR_LEVEL, 
                              "Bad state in server [%d]", state );
                     changeStatus( BADSTATE );
                     changeState( ERRORED );
                 }
                 else
                 {
                     // Set policy values
                     pktSize = pkVar.Value();
                     winSize = wVar.Value();

                     // Divide with pkt_size to find total_n_packets to be sent
                     totalPackets = entry->fileSize()/pktSize + 
                                    (entry->fileSize()%pktSize > 0);
                     totalBlocks  = totalPackets/winSize + 
                                    (totalPackets%winSize > 0);
    
                     // In group, start send announcement, set timer, set
                     // retries
                     sendAnnouncement();
                     fxrTimerReset( startTimer );
                     changeState( WAIT_START );
                     retries = 0;
                 }
             }
             break;

        case WAIT_START: // Waiting for joins/beginning (server)
             if ( startTimer.expired() == TRUE )
             {
                 // Increment retries, check for done
                 if ( ++retries >= START_RETRIES )
                 {
                     // Set init state, cleanup start timer
                     changeState( INIT_XFER );
                     waitTimers.removeTimer( &startTimer );
                 }
                 else
                 {
                     // Send announcement, set timer, set retries
                     sendAnnouncement();
                     fxrTimerReset( startTimer );
                 }
             }
             break;

        case INIT_XFER: // Initialize the transfer
             if ( initXfer() == TRUE )
             {
                 // Report transfer begin
                 LOGMSG5( log, FXFER_REPORT_LEVEL, "Starting transfer "
                          "[%s,packets=%d(sz=%d),blocks=%d(ws=%d)]",
                          (char *)*entry, totalPackets, pktSize, 
                          totalBlocks, winSize );

                 // Start the transfer
                 nextBlock( 0 );
                 inProgress = TRUE;
                 changeState( SEND_DATA );
             }
             break;
             
        case SEND_DATA: // Send the data
             sendData();
             fxrTimerReset( ackTimer );
             changeState( XFER_WAIT );
             break;

        case XFER_WAIT: // Wait for incoming data
             // If an ack message was received
             if ( mt == DATA_ACK )
             {
                 // Process the acknowledgement
                 processAcknowledgement( msg );

                 // Check if everyone has acked
                 if ( ackQueue.listSize() == 0 )
                 {
                     // Check if everything acked
                     if ( sendMask->onBits() == 0 )
                     {
                         // Start next block
                         if ( blockCounter+1 >= totalBlocks )
                         {
                             // End the transmission
                             sendFin();
                             fxrTimerReset( finTimer );
                             changeState( WAIT_END );
                             retries = 0;
                             finishFile( TRUE );

                             // Cleanup ack timer
                             waitTimers.removeTimer( &ackTimer );
                         }
                         else
                         {
                             // Proceed to next block
                             nextBlock( blockCounter+1 );
                             changeState( SEND_DATA );
                         }
                     }
                     else
                     {
                         // Add everyone who is not fully acknowledged
                         AMXferMem *mem;
                         xferQueue.resetIterator();
                         while ( (mem = (AMXferMem *)xferQueue.
                                                      nextIterator()) != NULL )
                         {
                             // Reset ack mask, insert into acknowledge list
                             if ( (ackQueue.find(mem) == NULL) && 
                                  (mem->allAcked() == FALSE) )
                             {
                                 // Add them back into ack list
                                 mem->reset();
                                 ackQueue.insertObj( mem );
                             }
                         }

                         // Xmit what remains unacknowledged
                         retries = 0;
                         changeState( SEND_DATA );
                     }
                 }
             }
             // If a timeout has occured
             else if ( ackTimer.expired() == TRUE )
             {
                 // Check if we have waited long enough
                 if ( ++retries >= ACK_RETRIES )
                 {
                     // Purge members from the group who have not responded
                     if ( purgeAckers() == FALSE )
                         return( FALSE );

                     // Start next block
                     if ( blockCounter+1 >= totalBlocks )
                     {
                         // End the transmission
                         sendFin();
                         fxrTimerReset( finTimer );
                         changeState( WAIT_END );
                         retries = 0;

                         // Cleanup ack timer
                         waitTimers.removeTimer( &ackTimer );
                     }
                     else
                     {
                         // Proceed to next block
                         nextBlock( blockCounter+1 );
                         changeState( SEND_DATA );
                     }
                 }
                 else
                 {
                     // Mark last packet to be re-sent, send data
                     sendMask->Set( blkSize-1, TRUE );
                     changeState( SEND_DATA );
                     fxrTimerReset( ackTimer );
                 }
             }
             break;

        case WAIT_END: // End of transfer
             // If a timeout has occured
             if ( finTimer.expired() == TRUE )
             {
                 // Check if we have waited long enough (or group empty)
                 if ( (++retries >= FIN_RETRIES) || (group->groupSize() <= 1) )
                 {
                     // Log finish
                     LOGMSG1( log, FXFER_REPORT_LEVEL, "Completed xfer [%s]",
                              (char *)file );

                     // We are done, cleanup FIN timer
                     changeState( GRP_LEAVE );
                     waitTimers.removeTimer( &finTimer );

                     // Quit the group (non-blocking)
                     group->Quit( FALSE );
                 }
                 else
                 {
                     // Send the next fin
                     fxrTimerReset( finTimer );
                     sendFin();
                 }
             }
             break;

        case GRP_LEAVE: // Wait to leave group
             if ( group->State() != AGroup::AG_INOPERATION )
             {
                 // Log leave group, close up shop
                 LOGMSG1( log, FXFER_DETAIL_LEVEL, "Exited from group [%s]",
                          (char *)file );
                 changeStatus( XFERCOMPL );
             }
             break;

        default: // Unknown
             LOGMSG1( log, FXFER_ERROR_LEVEL, "Bad state in server [%d]", state );
             changeStatus( BADSTATE );
             changeState( ERRORED );
             break;
    }

    // Cleanup message
    if ( msg != NULL ) 
    { 
        // Log deletion
        LOGMSG2( log, FXFER_DETAIL_LEVEL, "Delete recv [%s,%p]",
                 messageString((MessageTypes)mt), msg );
        delete msg;
    }


    // Return successfully (or not)
    return( (status == OK) || (status == XFERCOMPL) );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::clientStateMachine
// Description  : execute the client state machine
//
// Inputs       : msg - message received, if any
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::clientStateMachine( Buffer * msg )
{
    // If message received
    unsigned short mt = MAX_MSGS;
    if ( msg != NULL )
    {
        // Get message type, log it
        *msg >> mt;
        LOGMSG1( log, FXFER_DETAIL_LEVEL, "Client recv [%s]", 
                 messageString((MessageTypes)mt) );
    }

    if (globalCheckOnly)
    {
        switch (state)
        {
            case START: // waiting for message from server
                if ( group->State() == AGroup::AG_INOPERATION )
                {
                    fxrTimerReset(startTimer);
                    changeState (WAIT_START);
                    AString *exportName = entry->Fsystem();
                    LOGMSG1(log, FXFER_DETAIL_LEVEL, "xfer group %s is active",
                        (char*)(*exportName));
                    printf ("++++ client reset startTimer\n");
                }
                break;

            case WAIT_START:
                if (mt == XFER_CHECK_REQ)
                {
                    printf ("++++ client got XFER_CHECK_REQ message\n");
                    printf ("++++ client group->LocalID(): %s\n",
                        (char*)(group->LocalID()));
                    // tell the server the status of our file
                    Buffer *buf = new Buffer;
                    Buffer fileHash;
                    if (entry->HashFile())
                    {
                        entry->GetFileHash (&fileHash);
                    }
                    else
                    {
                        fileHash << (uchar)0;
                    }
                    AString clientName = group->LocalID();
                    *buf << (unsigned short)XFER_CHECK << clientName << fileHash;
                    AString fileHashText = "";
                    fileHashText.Bin2Ascii ((char*)fileHash.Buf(), fileHash.Length());
                    LOGMSG4 (log, FXFER_DETAIL_LEVEL,
                        "%s sending check info for %s : %s   hash:%s",
                        (char*)clientName, (char*)(entry->Fsystem()),
                        (char*)(*entry), (char*)fileHashText);
                    if (sendInterface(buf))
                    {
                        changeState(GRP_LEAVE);
                    }
                }
                else
                {
                    if (startTimer.expired())
                    {
                        changeState(ERRORED);
                        changeStatus(NOCLIENT);
                        // Quit the group (non-blocking)
                        group->Quit( FALSE );
                        LOGMSG1( log, FXFER_ERROR_LEVEL,
                            "No server detected for checking file [%s]", (char*)file);
                        return FALSE;
                    }
                }
                break;

            case GRP_LEAVE: // Wait to leave group
                if ( group->State() != AGroup::AG_INOPERATION )
                {
                    // Log leave group, close up shop
                    LOGMSG1( log, FXFER_DETAIL_LEVEL,
                        "Exited from check group [%s]", (char *)file );
                    changeStatus(XFERCOMPL);
                }
            break;

            default:
                LOGMSG1( log, FXFER_ERROR_LEVEL,
                    "Bad state in check client [%d]", state );
                changeStatus( BADSTATE );
                changeState( ERRORED );
                break;
        }
        return (status == OK) || (status == XFERCOMPL);
    }

    // Switch on client state
    switch( state )
    {
        case START: // Initial state
             if ( group->State() == AGroup::AG_INOPERATION )
             {
                 // In group, start transfer
                 fxrTimerReset( startTimer );
                 changeState( WAIT_START );
             }
             break;

        case WAIT_START: // Waiting for joins/beginning (server)
             if ( mt == START_ANNOUNCE )
             {
                 // If good announcement
                 if ( processAnnouncement( msg ) == TRUE )
                     changeState( INIT_XFER );

                 // Report transfer begin
                 LOGMSG5( log, FXFER_REPORT_LEVEL, "Starting transfer "
                          "[%s,packets=%d(sz=%d),blocks=%d(ws=%d)]",
                          (char *)*entry, totalPackets, pktSize, 
                          totalBlocks, winSize );
             }
             // Never got an announcement bail
             else if ( startTimer.expired() == TRUE )
             {
                 // Log, error out
                 LOGMSG1( log, FXFER_ERROR_LEVEL, "No server detected [%s]",
                         (char *)file );
                 changeState( ERRORED );
                 changeStatus( NOCLIENT );
                 return( FALSE );
             }
             break;

        case INIT_XFER: // Initialize the transfer
             if ( initXfer() != FALSE ) 
             {
                 changeState( XFER_WAIT );
                 fxrTimerReset( startTimer );
             }
             break;

        case XFER_WAIT: // Wait for incoming data
             // Data received
             if ( mt == DATA_BLOCK )
             {
                 // Start timing from first block of data
                 if ( dTime == 0 ) dTime = AtkClock::HwClock();

                 // If not already in progress
                 if ( inProgress == FALSE )
                 {
                     // setup ack timer, current block
                     cnextBlock( 0 );
                     waitTimers.removeTimer( &startTimer );

                     // Set to "in progress"
                     inProgress = TRUE;
                 }

                 // Process the data packet
                 processData( msg );
                 fxrTimerReset( ackTimer );
                 retries = 0;
             }
             // Transfer completed
             else if ( mt == XFER_FIN )
             {
                 // Are we done?
                 if ( (blockCounter+1 == totalBlocks) && 
                      (sendMask->onBits() == sendMask->Length()) )
                 {
                     // Log completion
                     LOGMSG1( log, FXFER_REPORT_LEVEL, "Completed xfer [%s]",
                              (char *)file );

                     // We are done, cleanup FIN timer
                     changeState( GRP_LEAVE );
                     waitTimers.removeTimer( &finTimer );

                     // Quit the group (non-blocking)
                     group->Quit( FALSE );

                     // Cleanup file
                     if (writeNextBlock())
                     {
                         finishFile( TRUE );
                     }
                     else
                     {
                         finishFile( FALSE );
                         return FALSE;
                     }

                     // Figure out how long the download took
                     unsigned long dt = 
                         AtkClock::HwClockDiff( AtkClock::HwClock(), dTime );
                     LOGMSG3( log, FXFER_REPORT_LEVEL, "Download time [%s] "
                              "(%lu,%lu usecs/size)", (char *)file, dt, 
                              entry->fileSize() );
                 }
                 else
                 {
                     // Failed to finish transfer
                     LOGMSG1( log, FXFER_ERROR_LEVEL, "Premature FIN [%s]",
                             (char *)file );
                     changeState( ERRORED );
                     changeStatus( BADSTATE );
                     finishFile( FALSE );
                     return( FALSE );
                 }
             }
             // If still waiting for first packet and expired
             else if ( (inProgress == FALSE) && (startTimer.expired() == TRUE) )
             {
                 // Failed to get first packet, bail
                 LOGMSG1( log, FXFER_ERROR_LEVEL, "No first packet [%s]",
                         (char *)file );
                 changeState( ERRORED );
                 changeStatus( TOOMANYTO );
                 return( FALSE );
             }
             // Waited too long for xmit/next block
             else if ( (inProgress == TRUE) && (ackTimer.expired() == TRUE) )
             {
                 // Check if we have waited long enough
                 if ( ++retries >= ACK_RETRIES )
                 {
                     // Failed to get mext packet, bail
                     LOGMSG1( log, FXFER_ERROR_LEVEL, "Lost sync [%s]",
                             (char *)file );
                     changeState( ERRORED );
                     changeStatus( TOOMANYTO );
                     return( FALSE );
                 }
                 else
                 {
                     // Just acknowledge what we have
                     sendAcknowledgement();
                     fxrTimerReset( ackTimer );
                 }
             }
             break;

        case GRP_LEAVE: // Wait to leave group
             if ( group->State() != AGroup::AG_INOPERATION )
             {
                 // Log leave group, close up shop
                 LOGMSG1( log, FXFER_DETAIL_LEVEL, "Exited from group [%s]",
                          (char *)file );
                 changeStatus( XFERCOMPL );
             }
             break;

        default: // Unknown
             LOGMSG1( log, FXFER_ERROR_LEVEL, "Bad state in client [%d]", state );
             changeStatus( BADSTATE );
             changeState( ERRORED );
             break;
    }

    // Cleanup message
    if ( msg != NULL ) delete msg;

    // Return successfully
    return( (status == OK) || (status == XFERCOMPL) );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::finishFile
// Description  : take completed download, move to target, set attributes
//
// Inputs       : sucess - flag indicating success or failure of xmission
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::finishFile( Boolean success )
{
    // If already closed (or never opened), ignore
    if ( fhandle == -1 ) return( TRUE );

    // Close the file
    close( fhandle );
    fhandle = -1;

    // If client
    if ( group->IsServer() == FALSE )
    {
        // If the download has successfully concluded
        if ( success == TRUE )
        {
            // Unlink the old file
            unlink( (char *)file );
    
            // Move tempfile
            if ( link( (char *)tfile, (char *)file ) != 0 )
            {
                // Error out
                LOGMSG3( log, FXFER_ERROR_LEVEL, "Unable to move tempfile "
                         "[%s<-%s](%s)", (char *)file, (char *)tfile, 
                         strerror(errno) );
                changeState( ERRORED );
                changeStatus( GROUPERROR );
                return( FALSE );
            }

            // Set the mode and access time
            if ( entry->setFileInfo() == FALSE )
            {
                // Error out
                changeState( ERRORED );
                changeStatus( GROUPERROR );
                return( FALSE );
            }
        }

        // Unlink the download file
        unlink( (char *)tfile );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::sendInterface
// Description  : send a message to the group
//
// Inputs       : buf - buffer to send
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::sendInterface( Buffer * buf )
{
#ifdef DROP_TEST
#define DROP_RATE 5 // Percent
    // Drop a percentage of the packets
    if ( Random::GenerateRandom(0, 99) < DROP_RATE )
    {
        // Cleanup buffer, return successfully
        delete buf;
        return( TRUE );
    }
#endif

    // save the message type for a possible error message
    ushort ms = *(ushort*)buf->Buf();
    MessageTypes m = (MessageTypes)ms;
    char *msgName = (char*)messageString((MessageTypes)ntohs(*(unsigned short *)buf->Buf()));

    LOGMSG1( log, AMIRD_DETAIL_LEVEL, "AMirD Xfer Group Send [%s]", msgName);

    // Send the message
    group->sendMessage( buf );
    if ( group->State() != AGroup::AG_INOPERATION )
    {
        // Error out
        LOGMSG2( log, FXFER_ERROR_LEVEL,
            "[%s] send failed due to bad group state %s", 
            messageString(m), AGroup::StateTypeDescr(group->State()) );
        changeState( ERRORED );
        changeStatus( GROUPERROR );

        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::recvInterface
// Description  : send a message to the group
//
// Inputs       : buf - buffer to send
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::recvInterface( Buffer **buf )
{
    // Read the message, check for error state
    struct timeval tm = { 0, 0 };
    group->readMessage( buf, &tm );
    if ( group->State() == AGroup::AG_ERRORED ) 
    {
        // Error out
        LOGMSG0( log, FXFER_ERROR_LEVEL, "xfer recv failed" );
        changeState( ERRORED );
        changeStatus( GROUPERROR );
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//
// Message Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::sendAnnouncement
// Description  : send an announcment
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::sendAnnouncement( void )
{
    // Local variables
    AString pth;
    Buffer *buf = new Buffer();

    // Create announcement message (file/xfer info)
    *buf << (unsigned short)START_ANNOUNCE 
         << entry->relativePath(pth)
         << (unsigned char)entry->entryType();
    entry->doAnnounce( *buf, NULL );
    *buf << pktSize << winSize;

    // Log send
    LOGMSG4( log, FXFER_DETAIL_LEVEL, "Sending FXFER announcement [%s,%s,"
             "%ld,%ld]", (char *)entry->relativePath(pth), 
             (char *)*entry->Fsystem(), pktSize, winSize );

    // Send the message
    if ( sendInterface(buf) == FALSE ) return( FALSE );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::processAnnouncement
// Description  : recv an announcment
//
// Inputs       : msg - buffer containing message contents
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::processAnnouncement( Buffer * msg )
{
    // Get the message contents
    AString fname, relpath;
    unsigned char ty;
    Buffer summary;
    *msg >> fname >> ty;
    if ( ((entry->Fsystem() != NULL) && 
          (fname != entry->relativePath(relpath))) || (ty != AMEntry::ISFILE) ||
         (entry->reconcileAnnounce(*msg) == FALSE) )
    {

        // Why is this a bad announcement
        if ( (entry != NULL) && (entry != NULL) && 
             (fname != entry->relativePath(relpath)) )
        {
            // Incorrect group announcement, bail out
            LOGMSG2( log, FXFER_ERROR_LEVEL, "Incorrect group announce [%s!=%s]", 
                     (char *)file, (char *)relpath );
        }

        // Log the bad annoucement
        LOGMSG1( log, FXFER_ERROR_LEVEL, "Bad announcement [%s]", (char *)file );
        if ( entry == NULL )
        {
            LOGMSG0( log, FXFER_ERROR_LEVEL, "FS (null entry)" );
        }
        else if ( entry->Fsystem() == NULL )
        {
            LOGMSG0( log, FXFER_ERROR_LEVEL, "FS (null fs)" );
        }
        else
        {
            LOGMSG1( log, FXFER_ERROR_LEVEL, "FS %s", (char *)
                     (AString &)*entry->Fsystem() );
        }
        LOGMSG1( log, FXFER_ERROR_LEVEL, "FL %s", (char *)relpath );
        LOGMSG1( log, FXFER_ERROR_LEVEL, "Fi %s", (char *)fname );
        LOGMSG1( log, FXFER_ERROR_LEVEL, "FD %s", (char *)file );
        LOGMSG1( log, FXFER_ERROR_LEVEL, "ty %d", ty );

        // Set state to errored
        changeState( ERRORED );
        changeStatus( UNIXERROR );
        return( FALSE );
    }

    // Get xfer info, Divide with pkt_size to find total_n_packets to be sent
    *msg >> pktSize >> winSize;
    totalPackets = entry->fileSize()/pktSize + (entry->fileSize()%pktSize > 0);
    totalBlocks  = totalPackets/winSize + (totalPackets%winSize > 0);

    // Log the annoucement
    LOGMSG4( log, FXFER_DETAIL_LEVEL, "Received XFER announcement [%s,sz=%lu,"
             "pkt=%d,win=%d]", (char *)entry->localPath(relpath), 
             entry->fileSize(), pktSize, winSize );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::sendData
// Description  : send all unacknowledged data
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::sendData( void )
{
    // Walk list of read data
    for ( unsigned long i=0; i<winSize; i++ ) 
    {
        // If unacknowledged
        unsigned long pkno = (blockCounter*winSize)+i;
        if ( (sendMask->Value(i) == TRUE) && (pkno < totalPackets) )
        {
            // Create block, send data
            Buffer * msg = new Buffer();
            *msg << (unsigned short)DATA_BLOCK << blockCounter 
                 << i << *dataBufs[i];
            LOGMSG2( log, FXFER_DETAIL_LEVEL, "Send data "
                     "[blk=%d,pkt=%d]", blockCounter, i );
            if ( sendInterface(msg) == FALSE ) return( FALSE );
        }
    }

    // Set sent
    sendMask->SetVector( FALSE );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::processData
// Description  : Process the data packet
//
// Inputs       : msg - buffer containing message contents
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::processData( Buffer * msg )
{
    // First, get the data information
    unsigned long blk, pkt, wpkt = lastUnacked();
    *msg >> blk >> pkt;

    // Is this the first packet of a new block
    if ( blk == blockCounter+1 )
    {
        // Do we not have everything from this block
        if ( sendMask->onBits() != sendMask->Length() )
        {
            // Error out
            LOGMSG2( log, FXFER_ERROR_LEVEL, "Bad data recv "
                     "[blk=%d,pkt=%d]", blk, pkt );
            changeState( ERRORED );
            changeStatus( GROUPERROR );
            return( FALSE );
        }

        // OK, get next block
        cnextBlock( blk );
    }

    // If this is for the current block (and not already received)
    if ( (blk == blockCounter) && (sendMask->Value(pkt) == FALSE) )
    {
        // Read data, set mask, log
        *msg >> *dataBufs[pkt];
        sendMask->Set( pkt, TRUE );
        LOGMSG3( log, FXFER_DETAIL_LEVEL, "Received data for %s "
                 "[blk=%d,pkt=%d]", (char *)file, blk, pkt );
    }
    else
    {
        // Log repeat/bad data
        LOGMSG3( log, FXFER_DETAIL_LEVEL, "Ignoring data for %s "
                 "[blk=%d,pkt=%d]", (char *)file, blk, pkt );
    }

    // What is the last packet I am waiting on, is this it?
    // Last packet in window is implicit signal for acks
    if ( (pkt == blkSize-1) || (pkt == wpkt) )
    {
        // ack everything
        sendAcknowledgement();
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::sendAcknowledgement
// Description  : send an acknowldegement
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::sendAcknowledgement( void )
{
    // Create announcement message, send it
    Buffer *buf = new Buffer();
    *buf << (unsigned short)DATA_ACK << group->LocalID() << blockCounter;
    sendMask->toBuffer(*buf);
    if ( sendInterface(buf) == FALSE ) return( FALSE );

    // Log send 
    AString s;
    LOGMSG2( log, FXFER_DETAIL_LEVEL, "Sending acknowledgement [%d,%s]", 
             blockCounter, (char *)sendMask->vectorString(s) );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::processAcknowledgement
// Description  : process an incoming data block acknowledgement
//
// Inputs       : msg - buffer containing message contents
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::processAcknowledgement( Buffer * msg )
{
    // Local vars
    AMXferMem *mem;
    BitVector locVec(winSize), l2(winSize);
    AString id, str, str2;
    unsigned long blk;

    // Get the message contents
    *msg >> id >> blk;
    locVec.fromBuffer(*msg);

    // Check if relevant
    if ( (blk == blockCounter) && 
         ((mem = (AMXferMem *)ackQueue.deleteObj((char *)id)) != NULL) )
    {
        // Record any unacknowledged mask, set send mask
        mem->acknowledge( locVec );
        l2 = *mem;
        sendMask->OR( l2.COMP() );

        // Log the acknowledgement
        LOGMSG4( log, FXFER_DETAIL_LEVEL, "Received acknowledgement from %s "
                 "[%d,%s] sm [%s]",(char *)id, blk, 
                 (char *)locVec.vectorString(str),
                 (char *)sendMask->vectorString(str2) );
    }
    else
    {
       // Log strange ack
       LOGMSG3( log, FXFER_DETAIL_LEVEL, "Strange acknowledgement from %s [%d,%s]",
                (char *)id, blk, (char *)locVec.vectorString(str) );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMFxfer::sendFin
// Description  : send a transfer FIN message
//
// Inputs       : msg - buffer containing message contents
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMFxfer::sendFin( void )
{
    // Create FIN message, send it
    Buffer *buf = new Buffer();
    *buf << (unsigned short)XFER_FIN;
    if ( sendInterface(buf) == FALSE ) return( FALSE );

    // Log send
    LOGMSG0( log, FXFER_DETAIL_LEVEL, "Sending FIN" );

    // Return successfully
    return( TRUE );
}

