#ifndef AMFxfer_INCLUDED
#define AMFxfer_INCLUDED

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMFxfer.h
//
//   Created : 09-Oct-00
//   By : Patrick McDaniel
//
//   Description : This class embodies the file transfer protocol.  It
//                 creates and operates the protocol completely.
//
//   *** Change History ***
//
//   09-Oct-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// *** Include Files ***
#include <list>
#include <AGroup.h>

// Project Include Files
#include <AMFile.h>

// *** Definitions *** 
#define DEF_PACKET_SIZE 1024
#define DEF_WINDOW_SIZE 10

// Timer Values
#define START_WAIT_TO (4*1000) // 4 seconds
#define ACK_WAIT_TO (2*1000) // 2 seconds
#define FIN_WAIT_TO (2*1000) // 2 seconds

// Setup log levels
#define FXFER_ERROR_LEVEL  AMIRD_ERROR_LEVEL
#define FXFER_REPORT_LEVEL (AMIRD_DETAIL_LEVEL<<1)
#define FXFER_DETAIL_LEVEL (AMIRD_DETAIL_LEVEL<<2)
#define FXFER_REPORT (AMIRD_DETAIL_LEVEL<<3)

// Retries
#define START_RETRIES 3
#define ACK_RETRIES 5
#define FIN_RETRIES 3

// Policy Values
#define FX_PKT_SIZE "fxpktsz"
#define PSIZE_ATTR "psize"
#define FX_WIN_SIZE "fxwinsz"
#define WSIZE_ATTR "wsize"

// Forward References
class AMirDPolicyImpl;

// Class Definition AMFxfer
class AMFxfer : public Listable {

  public: // Public Type Definitions

    // Error state enumerations 
    typedef enum StatusEnum {
      OK          = 0,  // Everything OK
      TOOMANYTO   = 1,  // Too many timeouts
      ARGMISMATCH = 2,  // constructor for server side
      BADSTATE    = 3,  // Bad state detected
      UNIXERROR   = 4,  // UNIX (file) error detected
      NOCLIENT    = 5,  // No clients detected
      GROUPERROR  = 6,  // Group error
      XFERCOMPL   = 7,  // Transfer successfully completed
      MAX_STATE   = 8,  // Maximum state num
    } ProtocolStatus;

    // Progress states
    typedef enum ProtocolEnum {
      // PHASE 1
      START       = 0, // Initial state
      WAIT_START  = 1, // Waiting for joins/beginning

      // PHASE 2
      INIT_XFER   = 2, // Begin the transfer
      SEND_DATA   = 3, // Send data
      XFER_WAIT   = 4, // Wait for xfer event

      // PHASE 3
      WAIT_END    = 5, // Wait for final FIN packet/end
      GRP_LEAVE   = 6, // Wait for group leave

      // MISC
      ERRORED     = 7, // Error state
      WAIT_CHECK_REQ = 8, // server waiting for check
      MAX_PSTATE  = 9 // Maximum state number

    } ProtocolState;

    // Message types
    typedef enum MessageEnum {
      START_ANNOUNCE = 0, // Session Announcement
      DATA_BLOCK     = 1, // Data xfer block
      DATA_ACK       = 2, // Data acknowledegement
      XFER_FIN       = 3, // Transfer fin
      XFER_CHECK_REQ = 4, // server sends Request File status (only check file)
      XFER_CHECK     = 5, // client sends File status (only check file)
      MAX_MSGS       = 6  // Maximum message number
    } MessageTypes;

  public: // Public Interface

    //
    // Constructors and Destructors

    AMFxfer( ATransport * trans, AString & fl, AString & gpol, AString & lpol, 
             AuditLog * lg );
    // Server Constructor (testing, non filesystem)
    
    AMFxfer( ATransport * trans, AString & fl, AString & lpol, AuditLog * lg );
    // Client Constructor (testing, non filesystem)

    AMFxfer( ATransport * trans, AMFile *ent, AString & gpol, AString & lpol, 
             AuditLog * lg = NULL );
      // Server Constructor

    AMFxfer( ATransport * trans, AMFile *ent, AString & lpol, 
             AuditLog * lg = NULL );
      // Client Constructor

    virtual ~AMFxfer( void );
      // Destructor

    //
    // Access Methods

    // Return the current state of the xfer
    ProtocolStatus Status( void )
    {
        return( status );
    }

    // Return the current protocol phase
    ProtocolState ProtocolPhase( void )
    {
        return( state );
    }

    // Return the updated entry
    AMFile * Entry( void )
    {
        return( entry );
    }

    // Return a reference to the group
    AGroup * Group( void )
    {
        return( group );
    }

    // Return the transport object
    ATransport * Transport( void )
    {
        return( group->Transport() );
    }

    // 
    // Member Methods

    Boolean doXfer( fd_set * readfds = NULL );
      // Execute the file xfer state machine

    // Prepare select call fd_set structure (add appropriate file
    // descriptors)
    void fdSet( int &n, fd_set *rfds, fd_set *wfds, fd_set *efds, 
                struct timeval *tmr )
    {
        // Prepare select call fd_set structure (add appropriate file
        // descriptors)
        group->fdSet( n, rfds, wfds, efds, tmr );

        // Compare local timer queue against tmr
        long ne = waitTimers.nextExpirationIn(NULL);
        if ( (ne >= 0) && (ne < tmr->tv_sec*1000 + tmr->tv_usec/1000) )
        {
            // Set the new timer value
            tmr->tv_sec = ne/1000;
            tmr->tv_usec = (ne%1000)*1000;
        }
    }

    // Is there a message pending receive?
    Boolean messagePending( void )
    {
        return( group->messagePending() );
    }

    // Change status methods
    void changeStatus( ProtocolStatus st )
    {
        LOGMSG3( log, FXFER_DETAIL_LEVEL, "Status change [%s: %s->%s]", 
                 (char*)(*entry), statusString(status), statusString(st) );
        status = st;
    }

    // Change state methods
    void changeState( ProtocolState st )
    {
        LOGMSG3( log, FXFER_DETAIL_LEVEL, "State change [%s: %s->%s]", 
                 (char*)(*entry), pstatusString(state), pstatusString(st) );
        state = st;
    }

    // Is this subgroup errored 
    Boolean isErrored( void )
    {
        return( (Status() != OK) && (Status() != XFERCOMPL) );
    }

    // Has this subgroup completed the download
    Boolean isCompleted( void )
    {
        return( Status() == XFERCOMPL );
    }

    //
    // Listable Functions

    // Get a hash value of the object
    virtual unsigned short hashValue( void )
    {
        return( file.hashValue() );
    }

    // Compare this object to one like it
    // Returns : < 0 if local less than object
    //           = 0 if local equal to object
    //           > 0 if local greater than object
    // NOTE: Just compares the pointers
    virtual int compare( Listable *obj )
    {
       // Use entity compare
       return( file.compare(obj) );
    }

    // This object compares a listable value/identifier to the local
    // object.
    // Returns : < 0 if local less than object
    //           = 0 if local equal to object
    //           > 0 if local greater than object
    virtual int compareValue( void *val )
    {
       // Use entity compareValue
       return( file.compareValue(val) );
    }

    // Return a string describing the object
    virtual const char * toString( void )
    {
        static AString str;
        str.sprintf( "%s [%s]", (char *)file, 
                     (char *)group->Transport()->toString() );
        return( (char *)str );
    }

    // 
    // Static Member Methods

    // Return a string describing the current state
    static const char * statusString( ProtocolStatus st )
    {
        if ( (st < OK) || (st >= MAX_STATE) ) return( "Unknown" );
        return( statusStrings[st] );
    }

    // Return a string describing the current state
    static const char * pstatusString( ProtocolState st )
    {
        if ( (st < START) || (st >= MAX_PSTATE) ) return( "Unknown" );
        return( pstatusStrings[st] );
    }

    // Return a string describing the message
    static const char * messageString( MessageTypes m )
    {
        if ( (m < START_ANNOUNCE) || (m >= MAX_MSGS) ) return( "Unknown" );
        return( messageStrings[m] );
    }

  private: // Private Interface and Data

    //
    // Private Methods
    
    Boolean initGroup( ATransport * trans, AString * gp, AString * lp );
      // initialize the xfer group

    Boolean initXfer( void );
      // initialize the transfer

    Boolean nextBlock( unsigned long blk );
      // prepare the next block

    Boolean cnextBlock( unsigned long blk );
      // client prepare the next block

    Boolean readNextBlock( void );
      // Read next block from file

    Boolean writeNextBlock( void );
      // Write next block to file

    Boolean purgeAckers( void );
      // Purge those people who have not correctly acked

    Boolean serverStateMachine( Buffer * msg );
      // execute the server state machine
    
    Boolean clientStateMachine( Buffer * msg );
      // execute the client state machine

    Boolean finishFile( Boolean success );
      // take completed download, move to target, set attributes

    //
    // Message Methods

    Boolean sendInterface( Buffer * buf );
      // send a message to the group

    Boolean recvInterface( Buffer **buf );
      // send a message to the group

    Boolean sendAnnouncement( void );
      // execute the client state machine

    Boolean processAnnouncement( Buffer * msg );
      // recv an announcment

    Boolean sendData( void );
      // send all unacknowledged data

    Boolean processData( Buffer * msg );
      // Process the data packet

    Boolean sendAcknowledgement( void );
      // send an acknowldegement

    Boolean processAcknowledgement( Buffer * msg );
      // process an incoming data block acknowledgement

    Boolean sendFin( void );
      // send a transfer FIN message

    //
    // Utility Methods

    // Return that we are in a stable state (i.e., wait state)
    Boolean stableState( void )
    {
        return( (status == XFERCOMPL) || (state == START) ||
                (state == WAIT_START) || (state == XFER_WAIT) ||
                (state == WAIT_END) || (state == ERRORED) );
    }

    // Reset an timer
    void fxrTimerReset( AtkTimer & tmr )
    {
        tmr.reset(); // Reset, insert acknowledge timer
        waitTimers.removeTimer( &tmr );
        waitTimers.addTimer( &tmr );
    }

    // Get the packet size
    int packetSize( unsigned pkno )
    {
        if ((pkno+1 >= totalPackets) && (entry->fileSize()%pktSize > 0)) 
            return( entry->fileSize()%pktSize );
        return( pktSize );
    }

    // Figure out what packet we are waiting on (last unacked)
    unsigned long lastUnacked( void )
    {
        unsigned long wpkt = blkSize-1;
        while ( (wpkt > 0) && (sendMask->Value(wpkt) == TRUE) ) wpkt--;
        if ( (wpkt == 0) && (sendMask->Value(wpkt) == TRUE) ) wpkt = blkSize-1;
        return( wpkt );
    }

    // 
    // Class Attributes

    AGroup * group;
      // This is the group used to transfer the file.
   
    AMirDPolicyImpl *polImpl;
      // This the download group policy implementor

    AAttributeSet attrs;
      // This is the attribute set for the group.

    AAttribute fileAttr;
      // This is the file attribute for the group

    AAttribute authAttr;
      // This is the authority attribute for the group

    AAttribute authIdAttr;
      // This is the authority id attribute for the group

    AAttribute pSize;
      // This is the packet size attribute.

    AAttribute wSize;
      // This is the window size attribute.

    AuditLog * log;
      // This is the log to write debug information.

    AString file, tfile;
      // These are the (local) path and (xfer) file (and tempfile).

    AMFile *entry;
      // This is the information about the file.

    ProtocolStatus status;
      // This is the current status of the file xfer.

    // File Information

    unsigned long pktSize, winSize, totalPackets, totalBlocks, blkSize, dTime;
      // Statistical information about the xfer

    //
    // Protocol Progress State

    ProtocolState state;
      // This is the state of the protocol.

    AtkTimerQueue waitTimers;
      // This is a queue of timers we are waiting for.
 
    AtkTimer startTimer, ackTimer, finTimer;
      // This timer is used to wait for start, acknowledgements, fin

    unsigned long retries, blockCounter;
      // This is a number of retries for a particular operation.

    BitVector *sendMask;
      // These are vector marking what is sent, recved, and acked.
    
    Queue xferQueue, ackQueue;
      // This is the queue of members and unacked.
         
    Buffer ** dataBufs;
      // This is a list of the data buffers.
    
    int fhandle;
      // This is the file handle for the xfer file.

    Boolean inProgress;
      // Is the current transfer in progress.

    //
    // Static Class Attributes

    static const char * statusStrings[MAX_STATE];
      // These are description strings for the xfer status.

    static const char * pstatusStrings[MAX_PSTATE];
      // These are description strings for the protocol state.
 
    static const char * messageStrings[MAX_MSGS];
      // These are description strings for the messages.

};

typedef AMFxfer* AMFxferPtr;
typedef list<AMFxferPtr> AMFxferList;

// End Class Definition AMFxfer

#endif
