/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMImpFsys.cpp
//
//   Created : 27-Dec-00
//   By : Patrick McDaniel
//
//   Description : (See AMImpFsys.h for class description)
//
//   *** Change History ***
//
//   27-Dec-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <Random.h>
#include <Queue.h>

// Project Include Files
#include <AMImpFsys.h>
#include <AMFile.h>

// Definitions

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMImpFsys::processUpdateRequest
// Description  : Process an incoming announcment
//
// Inputs       : ent - entry to delete
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMImpFsys::processUpdateRequest( Buffer * buf )
{
    // Get the file associated with this request
    AMFile *fl;
    if ( (fl = findFileById(buf)) != NULL )
    {
        // Reset req timer (If needed)
        if ( fl->UpReqTimer() != NULL ) fl->UpReqTimer()->reset();

        // Log, return processed here
        LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Recevied update req %s", 
                 (char *)*fl );
        return( TRUE );
    }

    // Return not processed here
    return( FALSE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMImpFsys::createFileId
// Description  : Create a file Id buffer
//
// Inputs       : fl - file to create request 
// Outputs      : buffer if successful, NULL otherwise

Buffer * AMImpFsys::createFileId( AMFile *fl )
{
    // Create buffer, stuff contents
    AString pth;
    Buffer * buf = new Buffer();
    (*buf) << Exporter() << exportedPath << fl->relativePath(pth);

    // Return the created buffer
    return( buf );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMImpFsys::findFileById
// Description  : Find the file associated with this ID
//
// Inputs       : fl - file to create request 
// Outputs      : file if successful, NULL otherwise (not for local fs)

AMFile * AMImpFsys::findFileById( Buffer *buf )
{
    // Get the data out of the message
    unsigned short pos = buf->Position();
    AString ex, path, rpath;
    (*buf) >> ex >> path >> rpath;

    // If meant for this filesystem and found
    AMFile * fl = ((ex != Exporter()) || (path != exportedPath)) ? 
                  NULL : (AMFile *)findEntry( rpath );

    // If not found, reset position, return result
    if ( fl == NULL ) buf->Position() = pos;
    return( fl );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMImpFsys::processAnnouncement
// Description  : Process an incoming announcment
//
// Inputs       : buf - announcement buffer
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMImpFsys::processAnnouncement( Buffer & buf )
{
    // Get the header data out of the announcement
    AString exr, epath, rel;
    buf >> exr >> epath >> rel;

    // Check if relevant (for this import?)
    if ( (exr == Exporter()) && (epath == exportedPath) )
    {
        // Find the base directory object
        AMEntry * entry;
        if ( (entry  = findEntry( rel )) == NULL)
        {
            // Error out
            LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Unable to find entry [%s]", 
                     (char *)rel );
        }
        else if ( entry->entryType() != AMEntry::ISDIR )
        {
            // Error out
            LOGMSG1( Log(), AMIRD_ERROR_LEVEL, "Bad entry type in announce [%s]", 
                     (char *)rel );
        }
        else
        {
            // Log the processing
            LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Received update [%s]", 
                     (char *)entry->localPath(rel) );

            // Scan, then process announcement
            AMDir *dir = ((AMDir *)entry);
            dir->scanEntryInfo( FALSE ); // Non-recursive
            dir->processAnnouncement( buf );
        }

        // Return relevant
        return( TRUE );
    }

    // Reset buffer, return that this is not relevant.
    buf.Position() = 0;
    return( FALSE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMImpFsys::getUpdateRequests
// Description  : create a queue of update requests
//
// Inputs       : reqs - queue of update requests
//              : random - random number base
// Outputs      : none

void AMImpFsys::getUpdateRequests(BufferList *reqs, Random *random)
{
    // Walk through pending list
    AMFile *fl;
    PendingUpdates().resetIterator();
    while ( (fl = (AMFile *)PendingUpdates().nextIterator()) != NULL )
    {
        // Create a request if needed
        if ( (fl->explicitRequest() == TRUE) && 
             ((fl->UpReqTimer() == NULL) || (fl->UpReqTimer()->expired())) )
        {
            // Create the request
            reqs->push_back(createFileId(fl));

            // If not defined yet, create timer
            if ( fl->UpReqTimer() == NULL )
            {
                // Create randomized re-request timer (i.e., SRM style reqs)
                unsigned long inter = UPDATE_REQUEST_PERIOD +
                                      (10*random->GenerateRandom( 0, 100 ));
                fl->UpReqTimer() = new AtkTimer( inter, fl );
            }

            // Reset the timer, log
            fl->UpReqTimer()->reset();
            LOGMSG1( Log(), AMIRD_DETAIL_LEVEL, "Update request created [%s]",
                     (char *)*fl );
        }
    }
}
