/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMirD.cpp
//
//   Created : 25-Sep-00
//   By : Patrick McDaniel
//
//   Description : (See AMirD.h for class description)
//
//   *** Change History ***
//
//   25-Sep-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <errno.h>
#include <IPAddress.h>
#include <Random.h>

// Project Include Files
#include <AMirD.h>
#include <AMirDPolicyImpl.h>

// Definitions

//
// Static Class Data
const char * AMirD::messageStrings[] = {
  "AMIRD_FS_UNKNOWN", "AMIRD_FS_UPDATE", "AMIRD_FS_REQUEST",
  "AMIRD_FS_SUBGRP", "AMIRD_FS_DELETE", "AMIRD_FS_QUIT"
};

//
// Static Configurations
BooleanVar AMirD::followSymLinks( SYMLINKS_CONFIG, TRUE );
BooleanVar AMirD::lookbackMCast( LOOPBACK_CONFIG, FALSE );
IntVar AMirD::minUpdatePeriod( MINUPDATE_CONFIG );

Boolean globalCheckOnly = FALSE;  // if TRUE, do not change files, only check them

//
// Static Global Functions

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : checkTransportType
// Description  : test whether this represents a good transport type
//
// Inputs       : v - the string
// Outputs      : none

static void checkTransportType( const AString * v )
{
    // Check the transport type
    if ( ATransport::transportType( *((AString *)v) ) == ATransport::AT_UNKNOWN )
    {
        // Throw on error
        AString msg ("Bad transport : "); msg += *v;
        throw new AtkException ( (char *)msg );
    }

    // Return, no return code
    return;
}

//
// Constructors and Destructors

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::AMirD
// Description  : Default Constructor
//
// Inputs       : fname - configuration filename (and path)
// Outputs      : none

AMirD::AMirD( AString & fname, AString * logName = NULL )
  : log( NULL ),
    configFile( fname ),
    shutdown( FALSE ),
    reconfig( TRUE ),
    synchronize( FALSE ),
    downloadTime( 0 ),
    updateAnnounceTimer( INTER_ANNOUNCE_PERIOD, NULL, AMIRD_EXPORT_SND ),
    updateRequestTimer( INTER_REQUEST_PERIOD, NULL, AMIRD_IMPORT_REQ ),
    ctrlTrans( NULL ),
    ctrlGroup( NULL ),
    authCert( NULL ),
    polImpl( NULL ),
    portsInUse( HARD_MAX_SUBGROUPS ),
    grpPolicy( GRPPOL_CONFIG ),
    locPolicy( LOCPOL_CONFIG ),
    exporter( EXPORTER_CONFIG ),
    upPeriod( UPPERIOD_CONFIG, DEFAULT_UPDATE_PERIOD ),
    numPeriod( NUMPERIOD_CONFIG, DEFAULT_NUM_PERIOD ),
    checkOnly( CHECKONLY_CONFIG, DEFAULT_CHECK_ONLY ),
    isServer( ISSRVR_CONFIG ),
    maxExportSgroups( MAX_ESG_CONFIG ),
    maxImportSgroups( MAX_ISG_CONFIG ),
    authorityCert( AUTHOR_CERT ),
    transType( TRANSPT_CONFIG ),
    mcAddr( MCADDR_CONFIG ),
    mcPort( MCPORT_CONFIG ),
    srvrAddr( SRVADDR_CONFIG ),
    srvrPort( SRVPORT_CONFIG ),
    sgtransType( SGTRANSPT_CONFIG ),
    sgmcAddr( SGADDR_CONFIG ),
    sgmcPort( SGPORT_CONFIG ),
    sgsrvrAddr( SGSRVAD_CONFIG ),
    sgsrvrPort( SGSRVPT_CONFIG )
{
    // Setup configuration value objects
    StringVarCheckPtr tchk = checkTransportType;
    transType.setCheck( tchk );
    sgtransType.setCheck( tchk );

    // Zero the port use vector
    portsInUse.SetVector( FALSE );

    // If needed, do log
    if ( logName != NULL )
    {
        // Create log, set log levels
        log =  new AuditLog( (char *)*logName );
        log->addAuditLevel( AMIRD_ALEVEL );
    }

    // create log for reporting
    report = new AuditLog ("report");
    report->setTermEcho (FALSE);
    report->setLogEcho (TRUE);
    report->setTimestamping (FALSE);
    report->addAuditLevel (-1);
    LOGMSG0 (report, AMIRD_REPORT_LEVEL, "AMirD Summary Report");

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::~AMirD
// Description  : Destructor
//
// Inputs       : none
// Outputs      : none

AMirD::~AMirD( void )
{
    // Return, no return code
    return;
}

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::execute
// Description  : execute the application loop
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::execute( void )
{
    // Local variables
    struct timeval tm;
    long tmout;
    fd_set rfds;
    int n, retval;
    AMFxfer * sgroup;
    AtkTimer exSpawnTimer(500);

    // While not shutting down
    exSpawnTimer.reset();
    while ( shutdown == FALSE )
    {
        // If we need to reconfigure
        if ( reconfig == TRUE )
        {
            // Cleanup groups, (Re-)read the configuration file
            cleanup();
            if ( readConfigFile() == FALSE )
                return( FALSE ); // Bail on error
            reconfig = FALSE;

            // Initialize the control group
            if ( initializeControlGroup() == FALSE ) return( FALSE );
        }

        // If we need to synchronize the filesystem
        if (synchronize)
        {
            AMFsysList::iterator fsIter;

            // Walk list of export mirrored filesystems
            for (fsIter = exports.begin(); fsIter != exports.end(); fsIter++)
            {
                synchFS( *(*fsIter) );
            }

            // Walk list of import mirrored filesystems (request)
            AMImpFsysList::iterator fsImpIter;
            for (fsImpIter = imports.begin(); fsImpIter != imports.end(); fsImpIter++)
            {
                sendSynchReq( *fsImpIter );
            }

            // Flag synchronization as complete, reset timer
            synchronize = FALSE;
        }

        // Process timer expirations
        processTimers(log);

        // Set timeout
        if ( (tmout = exQueue.nextExpirationIn(NULL)) == -1 )
        {
            // Default to 86400 seconds (24 hours)
            tm.tv_sec = MAX_AMIRD_WAIT;
            tm.tv_usec = 0;
        }
        else
        {
            // Set timeout explicitly
            tm.tv_sec = tmout/1000;
            tm.tv_usec = (tmout%1000)*1000;
        }

        // Setup File descriptor set
        n = 0;
        FD_ZERO( &rfds );
        if ( ctrlGroup != NULL )  
            ctrlGroup->fdSet( n, &rfds, NULL, NULL, &tm );

        // Walk the list of import and export groups
        AMFxferList::iterator imIt;
        for (imIt = importList.begin(); imIt != importList.end(); imIt++)
        {
            sgroup = *imIt;
            sgroup->fdSet( n, &rfds, NULL, NULL, &tm );
        }

        AMFxferList::iterator exIt;
        for (exIt = exportList.begin(); exIt != exportList.end(); exIt++)
        {
            sgroup = *exIt;
            sgroup->fdSet( n, &rfds, NULL, NULL, &tm );
        }


        // OK, now wait for until event occurs (check return code)
        retval = select( n, &rfds, NULL, NULL, &tm );
        if ( retval < 0 )
        {
            // Ignore signal interuptions
            if ( errno == EINTR )
                FD_ZERO( &rfds );
            else
            {
    
                // Error out
                LOGMSG1( log, AMIRD_ERROR_LEVEL, "Select failed [%s]", 
                         strerror(errno) );
                return( FALSE );
            }
        }

        // If the control group received a message, process it
        if ( ctrlGroup->isReadSet( &rfds ) == TRUE ) processMessage();

        // Work the import groups
        for (imIt = importList.begin(); imIt != importList.end(); imIt++)
        {
            sgroup = *imIt;
            // Work the import group
            sgroup->doXfer( &rfds );

            // If it has completed or errored, abort it
            if ( sgroup->Status() != AMFxfer::OK )
            {
                // Log shutdown
                LOGMSG2( log, AMIRD_REPORT_LEVEL, "Shutting down import group "
                         "[%s], state[%s]", (char *)sgroup->toString(),
                         AMFxfer::statusString(sgroup->Status()) );

                // Delete it and reset iterator
                shutdownImportSubgroup( sgroup );
                imIt = importList.begin();
            }
        }

        // Work the export groups
        for (exIt = exportList.begin(); exIt != exportList.end(); exIt++)
        {
            sgroup = *exIt;
            // Work the import group
            sgroup->doXfer( &rfds );

            // If it has completed or errored, abort it
            if ( sgroup->Status() != AMFxfer::OK )
            {
                // Log shutdown
                LOGMSG2( log, AMIRD_REPORT_LEVEL, "Shutting down export group "
                         "[%s], state[%s]", (char *)sgroup->toString(),
                         AMFxfer::statusString(sgroup->Status()) );

                // Delete it and reset iterator
                shutdownExportSubgroup( sgroup );
                exIt = exportList.begin();
            }
        }

        // Now create new export groups as necessary
        while ( (exportList.size() < (size_t)maxExportSgroups.Value()) &&
                 xfersPending(&exports) && 
                 exSpawnTimer.expired()
              )
        {
            // Spawn subgroup, reset timer
            spawnExportSubgroup();
            exSpawnTimer.reset();
        }

        // if all the work is done, then quit.
        if ((!xfersPending(&exports)) && (numPeriod.Value() == 0))
        {
            shutdown = TRUE;
            LOGMSG0( log, AMIRD_REPORT_LEVEL,
                "Last export group has completed.  AMird is shutting down.");
        }
    }

    if (isServer)
    {
        // tell the group to halt.
        sendQuit();
    }

    // Shutdown everything
    cleanup();

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::LogConfiguration
// Description  : Log the current configuration.
//
// Inputs       : lvl - log level to print config
// Outputs      : TRUE if successful, FALSE otherwise

void AMirD::LogConfiguration( unsigned long lvl )
{
    // Check if we need to bother
    if ( (log == NULL) || (!log->auditLevelOn(lvl)) ) return;

    // Print the file/dir information, return
    LOGMSG1( log, lvl, "** Configuration : Exporting for [%s]", 
             (char *)exporter.Value() );

    // Walk list of exported filesystems
    LOGMSG0( log, lvl, "** Exported filesystems" ); 
    AMFsys *fsys = NULL;
    AMFsysList::iterator iter;
    for (iter = exports.begin(); iter != exports.end(); iter++)
    {
        // Log the export
        fsys = *iter;
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "*   %s", (char *)*fsys );
    }

    // Walk list of imported filesystems
    LOGMSG0( log, lvl, "** Imported filesystems" ); 
    AMImpFsys *ifsys = NULL;
    AMImpFsysList::iterator impIter;
    for (impIter = imports.begin(); impIter != imports.end(); impIter++)
    {
        ifsys = *impIter;
        // Log import (and where mounted)
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "*   %s", (char *)ifsys->ExportPath() );
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "*     --> on %s", (char *)*ifsys );
    }
}

//
// Static Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::getAttributeParm
// Description  : Get get an attribute by policy
//
// Inputs       : grp - group object
//                nm - attribute name
//                var - variable
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::getAttributeParm( AGroup *grp, Var & var, 
                                 const char *def, AuditLog *lg )
{
    // Attempt to get the attribute, default to value
    AString val;
    if ( grp->AttributeSet()->getAttributeValue(var.Name(), val) == FALSE ) 
        val = def;

    // Try to parse out value
    try { var.parse( &val ); }
    catch( AtkException *e )
    {
        // Log, error out
        LOGMSG2( lg, FATAL_ERROR, "Bad parameter value for parameter [%s], "
                 "value [%s]", (char *)var.Name(), (char *)val );
        return( FALSE );
    }

    // Log acceptance, return succefully
    LOGMSG2( lg, AMIRD_DETAIL_LEVEL, "Accept parameter [%s], "
             "value [%s]", (char *)var.Name(), (char *)val );
    return( TRUE );
}

//
// Private Interfaces

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processTimers
// Description  : process all expired timers
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::processTimers( AuditLog *lg )
{
    // Process timer expirations
    AtkTimer * tmr;
    while ( (tmr = exQueue.getNextExpired()) != NULL )
    {
        // What kind of timer expired?
        switch( tmr->Type() )
        {
            case AMIRD_UPDATE_TIMER: // Send update announce/requests
                 if (numPeriod.Value() != 0)
                 {
                     synchFS( *(AMFsys *)tmr->timerObject() );
                     tmr->reset();
                     exQueue.addTimer( tmr );
                 }
                 if (numPeriod > 0)
                 {
                     numPeriod.setValue (numPeriod.Value() - 1) ;
                     LOGMSG1( lg, AMIRD_DETAIL_LEVEL,
                         "numperiod decremented to %d", numPeriod.Value());
                 }
                 break;
      
            case AMIRD_EXPORT_SND: // Space export announcements
                 sendUpdate();
                 break;

            case AMIRD_IMPORT_REQ: // Space import requests
                 sendUpdateRequest();
                 break;

            default: // Unknown type, error
                 LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unknown timer type [%d]",
                          tmr->Type() );
                 break;
        }
    }

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::readConfigFile
// Description  : Read the amird configuration file
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::readConfigFile( void )
{
    // Open the configuration file
    ifstream ifstr( (char *)configFile );

    // Reset configuration to default values
    exporter.setValue("");

    // If bad file
    if ( ! ifstr )
    {
        // Error out
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unable to read configuration "
                 "file [%s], aborting", (char *)configFile );
        return( FALSE );
    }

    // Read the lines of the configuration
    configState cfg = AMIRD_CFG_UNKNOWN;
    unsigned char instr[AMIRD_MAXLINE_LEN];
    unsigned short len;
    AString ln;
    Boolean cont;
    while ( (ifstr) && (!ifstr.eof()) )
    {
        // Get the next (possibly continued) line
        ln = "";
        cont = TRUE;
        while ( cont == TRUE )
        {
            // Get the line, length
            ifstr.getline( (char *)instr, AMIRD_MAXLINE_LEN );
            if ( (len = strlen((char *)instr)) > 0 )
            {
                // Add the line, checking for continue char
                if ( (cont = (instr[len-1] == '\\') ? TRUE : FALSE) == TRUE ) 
                    instr[len-1] = 0x0;
                ln += instr;
            }
            else 
                cont = FALSE;
        }

        // Read the line, check for valid (and non-comment) line, add
        // to list of directories
        if ( (ln.Length() > 0) && (ln.At(0) != '#') &&
             (parseConfigFileLine(cfg, ln) == FALSE) )
        {
            // Error out
            LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unable to parse or bad "
                     "configuration [%s], aborting", (char *)configFile );
            return( FALSE );
        }
    }

    // Now setup some default values
    AString ref;
    if ( exporter.Value() == "" ) 
        exporter.setValue( (char *)IPAddress::GetLocalHostname(ref) );

    // Walk list of exported filesystems
    AtkTimer *tmr;
    AMFsys *fsys = NULL;
    AMFsysList::iterator iter;
    for (iter = exports.begin(); iter != exports.end(); iter++)
    {
        fsys = *iter;
        // Set exporter, add update timer
        tmr = new AtkTimer(upPeriod.Value()*1000, fsys, AMIRD_UPDATE_TIMER);
        tmr->reset();
        exQueue.addTimer( tmr );
    }

    // Walk list of imported filesystems
    AMImpFsys *ifsys = NULL;
    AMImpFsysList::iterator impIter;
    for (impIter = imports.begin(); impIter != imports.end(); impIter++)
    {
        ifsys = *impIter;
        // Set exporter, add update timer
        tmr = new AtkTimer(upPeriod.Value()*1000, ifsys, AMIRD_UPDATE_TIMER);
        tmr->reset();
        exQueue.addTimer( tmr );
    }

    // If needed, load the authority certificate
    if ( (authorityCert.Updated() == TRUE) && ((authCert = Certificate::
          loadCertificate(authorityCert.Value(),isServer.Value())) == NULL) )
    {
        // Error out
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unable to acquire control "
                 "authority certificate [%s], aborting", 
                 (char *)authorityCert.Value() );
        return( FALSE );
    }

    // Log the configuratoin
    LogConfiguration( AMIRD_REPORT_LEVEL );

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::parseConfigFileLine
// Description  : parse one line of the configuration
//
// Inputs       : cfg - configuration state
//                ln - current line
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::parseConfigFileLine( configState &cfg, unsigned char * ln )
{
    // If defining new section
    AString line( ln );
    line.cleansString();
    if ( (line.At(0) == '[') && (line.At(line.Length()-1) == ']') )
    {
        if ( strcasecmp((char *)line, CONFIG_SECTION_TEXT) == 0 )
            cfg = AMIRD_CFG_VARS;
        else if ( strcasecmp((char *)line, EXPORTS_SECTION_TEXT) == 0 )
            cfg = AMIRD_CFG_EXPORTS;
        else if ( strcasecmp((char *)line, IMPORTS_SECTION_TEXT) == 0 )
            cfg = AMIRD_CFG_IMPORTS;
        else 
        {
            // Bad section, report and abort
            LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad section heading, aborting.", 
                     (char *)line );
            return( FALSE );
        }
    }
    else
    {
        // Process configuration line
        AString fs, mp, ex, gp, lp, at;
        switch ( cfg )
        {
            case AMIRD_CFG_UNKNOWN: // Unknown state, bad data
                 LOGMSG0( log, AMIRD_ERROR_LEVEL, "Configuration outside section "
                          "heading, aborting." );
                 LOGMSG1( log, AMIRD_ERROR_LEVEL, "Config [%s]", (char *)line );
                 return( FALSE );
      
            case AMIRD_CFG_VARS: // Setting configuration
                 if ( parseConfigLine(line) == FALSE )
                 {
                     // Error out
                     LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad config [%s], aborting",
                              (char *)line );
                     return( FALSE );
                 }
                 break;
      
            case AMIRD_CFG_EXPORTS: // Exporting filesystems
                 if ( parseExportLine(line,fs,gp,lp,at) == FALSE )
                 {
                     // Error out
                     LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad export [%s], aborting",
                              (char *)line );
                     return( FALSE );
                 }

                 // Check to see if filesystem exists
                 if ( AMEntry::exists((char *)fs) == FALSE )
                 {
                     // Error out
                     LOGMSG1( log, AMIRD_ERROR_LEVEL, "exporting non-existent"
                              "filesystem [%s], aborting", (char *)fs );
                     return( FALSE );
                 }

                 // Store, log export
                 exports.push_back( new AMFsys(exporter.Value(),fs,gp,lp,at,log) );
                 LOGMSG1( log, AMIRD_REPORT_LEVEL, "Exporting [%s]", (char *)fs );
                 break;
      
            case AMIRD_CFG_IMPORTS: // Importing filesystems
                 if ( parseImportLine(line,ex,fs,mp,lp,at) == FALSE )
                 {
                     // Error out
                     LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad import [%s], aborting",
                              (char *)line );
                     return( FALSE );
                 }

                 // Check to see if mount point exists
                 if ( AMEntry::exists((char *)mp) == FALSE )
                 {
                     // Error out
                     LOGMSG2( log, AMIRD_ERROR_LEVEL, "importing to non-existent"
                              "mount point [%s->%s], aborting", (char *)fs, 
                              (char *)mp );
                     return( FALSE );
                 }

                 // Setup imported filesystem, log it
                 imports.push_back( new AMImpFsys(ex,mp,fs,lp,at,log) );
                 LOGMSG2( log, AMIRD_REPORT_LEVEL, "Importing [%s], mount "
                          "point [%s]", (char *)fs, (char *)mp );
                 break;
      
            default: // Bad enum, 
                 LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad config state (%d)", cfg );
                 return( FALSE );
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::parseConfigLine
// Description  : parse a string containing a configuration value
//
// Inputs       : ln - line
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::parseConfigLine( AString & line )
{
    // Parse out import, mount point
    int pos;
    AString templine = line, type, value;

    // Get configuation type
    templine.cleansString();
    if ( (pos = templine.firstOccurrence('=')) != EOS )
        type = templine.subString( 0, pos-1 );
    else
        return( FALSE );

    // Get value
    value = templine.subString(pos+1);
    type.cleansString();
    value.cleansString();

    // Get value (error on no value)
    if ( value.Length() == 0 ) return( FALSE );

    // Try to parse out data
    try 
    {
        LOGMSG2 (log, AMIRD_DETAIL_LEVEL,
            "Parsing config type: %s  value: %s",
            (char*)type, (char*)value);
        // OK, now process the value by type
        if ( strcasecmp((char *)type, (char *)grpPolicy.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            grpPolicy.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)locPolicy.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            locPolicy.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)exporter.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            exporter.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)upPeriod.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            upPeriod.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)numPeriod.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            numPeriod.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)checkOnly.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            checkOnly.parse( (char *)value );
            globalCheckOnly = checkOnly.Value();
        }
        else if ( strcasecmp((char *)type, (char *)isServer.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            isServer.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)maxExportSgroups.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            maxExportSgroups.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)maxImportSgroups.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            maxImportSgroups.parse( (char *)value );
        }

        else if ( strcasecmp((char *)type, (char *)authorityCert.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            authorityCert.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)transType.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            transType.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)mcAddr.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            mcAddr.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)mcPort.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            mcPort.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)srvrAddr.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            srvrAddr.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)srvrPort.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            srvrPort.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)sgtransType.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            sgtransType.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)sgmcAddr.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            sgmcAddr.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)sgmcPort.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            sgmcPort.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)sgsrvrAddr.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            sgsrvrAddr.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)sgsrvrPort.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            sgsrvrPort.parse( (char *)value );
        }
        else if ( strcasecmp((char *)type, (char *)lookbackMCast.Name()) == 0 )
        {
            // Check single name, parse value
            if ( value.firstOccurrence(' ') != EOS ) return( FALSE );
            lookbackMCast.parse( (char *)value );
        }
        //
        // Unknown configuration
        else
        {
            // Error out
            LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unknown config [%s]", 
                     (char *)line );
            return( FALSE );
        }
    }
    catch (...)
    {
        // Log error
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Bad config value [%s]", 
                 (char *)value );
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::parseExportLine
// Description  : parse a string containing a export line
//
// Inputs       : line - input line
//                fs - filesystem
//                gp - group policy
//                lp - local policy
//                at - authorization certificate filename
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::parseExportLine( AString & line, AString & fs, AString & gp, 
                                AString & lp, AString & at )
{
    // Parse out export, policy
    int pos;
    AString temp( line );

    // Get the data out of the line
    if ( ((pos = line.Next( fs )) != EOS) &&
         ((pos = line.Next( gp, pos )) != EOS) &&
         ((pos = line.Next( lp, pos )) != EOS) )
        pos = line.Next( at, pos );

/*    // Get local policy
    templine.cleansString();
    if ( (pos = templine.lastOccurrence(' ')) == EOS )
        return( FALSE );
    lp = templine.subString( pos+1 );
    lp.cleansString();

    // Get mountpoint
    templine = templine.subString( 0, pos-1 );
    templine.cleansString();
    if ( (pos = templine.lastOccurrence(' ')) == EOS )
        return( FALSE );
    gp = templine.subString( pos+1 );
    gp.cleansString();

    // Get filesystem
    if ( (pos = templine.lastOccurrence(' ')) == EOS )
    {
    }
    else
    {
        fs = templine.subString( 0, pos-1 );
        fs.cleansString();
        at = "";
    }
*/

    // Return filesystem is good/bad
    return( (fs.Length()) && (gp.Length()) && (lp.Length()) );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::parseImportLine
// Description  : parse a string containing a import line
//
// Inputs       : ln - line
//                expt - exporter
//                fs - filesystem to export
//                mpt - mount point
//                lp - local policy
//                at - authorization certificate filename
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::parseImportLine( AString & line, AString & expt, AString & fs, 
                                AString & mpt, AString & lp, AString & at )
{
    // Parse out import, mount point, policy
    int pos, pos2;
    AString temp( line );

    // Get the data out of the line
    if ( ((pos = line.Next( temp )) != EOS) &&
         ((pos = line.Next( mpt, pos )) != EOS) &&
         ((pos = line.Next( lp, pos )) != EOS) )
        line.Next( at, pos );

    // Parse out exporter/filesystem
    if ( (pos2 = temp.Next( expt, 0, ':' )) != EOS )
        temp.Next( fs, pos2, ':' );

/*
    // Get local policy
    templine.cleansString();
    if ( (pos = templine.lastOccurrence(' ')) == EOS )
        return( FALSE );
    lp = templine.subString( pos+1 );
    lp.cleansString();

    // Get mountpoint
    templine = templine.subString( 0, pos-1 );
    templine.cleansString();
    if ( (pos = templine.lastOccurrence(' ')) == EOS )
        return( FALSE );
    mpt = templine.subString( pos+1 );
    mpt.cleansString();

    // Get exporter, filesystem
    templine = templine.subString( 0, pos-1 );
    if ( (pos = templine.firstOccurrence(':')) == EOS )
        return( FALSE );
    expt = templine.subString( 0, pos-1 );
    expt.cleansString();

    // Get filesystem
    fs = templine.subString( pos+1 );
    fs.cleansString();
*/

    // Return import is good/bad
    return( (mpt.Length()) && (expt.Length()) && (fs.Length()) && (lp.Length()) );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::findImport
// Description  : find an import associated with the export update
//
// Inputs       : fs - filesystem to find
// Outputs      : TRUE if successful, FALSE otherwise

AMFsys * AMirD::findImport( AMFsys & fs )
{
    // Walk list of imported filesystems

    AMImpFsys *ifsys = NULL;
    AMImpFsysList::iterator impIter;
    for (impIter = imports.begin(); impIter != imports.end(); impIter++)
    {
        ifsys = *impIter;
        if ( (AString &)fs == (AString &)*ifsys )
            return ifsys;
    }

    // Return "not found"
    return( NULL );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::cleanup
// Description  : cleanup groups and configuration
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

void AMirD::cleanup( void )
{
    // Cleanup xfer subgroups
    AMFxfer * sgroup;


    while (!importList.empty())
    {
        sgroup = importList.front();
        shutdownImportSubgroup( sgroup );
        importList.pop_front();
    }

    while (!exportList.empty())
    {
        sgroup = exportList.front();
        shutdownExportSubgroup( sgroup );
        exportList.pop_front();
    }

    // Shutdown all groups
    if ( ctrlGroup != NULL )
    {
        // Close groups and delete information
        ctrlGroup->Quit();
        delete ctrlGroup;
        ctrlGroup = NULL;
        delete ctrlTrans;
        ctrlTrans = NULL;
        delete polImpl;
        polImpl = NULL;
    }

    // Purge directory listings

    while (!exports.empty())
    {
        delete exports.front();
        exports.pop_front();
    }

    while (!imports.empty())
    {
        delete imports.front();
        imports.pop_front();
    }
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::initializeControlGroup
// Description  : initialize the control group 
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::initializeControlGroup( void )
{
    // Check the transport types, create transport
    AString m = transType.Value(), mca = mcAddr.Value(), sa = srvrAddr.Value();
    ctrlTrans = initTransport( ATransport::transportType(m), isServer.Value(), 
                               mca, mcPort.Value(), sa, srvrPort.Value() );
    if ( ctrlTrans == NULL ) return( FALSE );
    LOGMSG1( log, AMIRD_REPORT_LEVEL, "Connected to control group [%s]",
             (char *)ctrlTrans->toString() );

    // Attempt to start control group
    try
    {
        // Setup the attributes (if needed)
        if ( authCert != NULL )
        {
            // Setup the authority/id attributes
            StringVar *sv = new StringVar( "author", 
                                           (char *)authorityCert.Value() );
            authAttr.setAttribute( sv, TRUE );
            attrs.addAttribute( &authAttr );
            sv = new StringVar( "authorid", 
                                (char *)authCert->Subject().CommonName() );
            authIdAttr.setAttribute( sv, TRUE );
            attrs.addAttribute( &authIdAttr );
        }

        // Create the control group
        AString noCli("");
        AString gp = grpPolicy.Value(), lp = locPolicy.Value();
        polImpl = new AMirDPolicyImpl( this, log );
        if ( isServer.Value() == TRUE )
            ctrlGroup = new AGroup( ctrlTrans, gp, lp, noCli, polImpl, &attrs, log );
            // ctrlGroup = new AGroup( NULL, gp, lp, noCli, polImpl, &attrs, log );
        else
            ctrlGroup = new AGroup( ctrlTrans, lp, polImpl,  &attrs, log );
            // ctrlGroup = new AGroup( NULL, lp, polImpl,  &attrs, log );
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        LOGMSG1( log, FATAL_ERROR, "Exception create control group [%s]", 
                 e->Description() );
        return( FALSE );
    }

    // Connect/establish the group
    if ( ctrlGroup->Connect() != TRUE )
    {
        // error out
        LOGMSG0( log, FATAL_ERROR, "Unable to establish control context" );
        return( FALSE );
    }


    // Get the application policy
    if ( (getAttributeParm(ctrlGroup, maxExportSgroups, 
                           MAX_EXPORT_SUBGRPS, log) == FALSE) ||
         (getAttributeParm(ctrlGroup, maxImportSgroups, 
                           MAX_IMPORT_SUBGRPS, log) == FALSE) ||
         (getAttributeParm(ctrlGroup, followSymLinks, 
                            DEF_SYMLINK_ENABLE, log) == FALSE) ||
         (getAttributeParm(ctrlGroup, minUpdatePeriod, 
                           DEF_MIN_UP_PERIOD, log) == FALSE) )
    {
        // error out
        LOGMSG0( log, FATAL_ERROR, "Error applying application policy." );
        return( FALSE );
    }

    // Set the symbolic links, min allowable update period
    AMFsys::FollowSymbolicLinks() = followSymLinks.Value();
    AMFsys::MinUpdatePeriod() = minUpdatePeriod.Value();

    // Now a "special" policy allowing us to turn up/down the log level
    AString ll;
    if ( ctrlGroup->AttributeSet()->getAttributeValue("loglevel", ll) == TRUE )
    {
        // Convert/set the log level
        IntVar llevel( "loglevel" );
        llevel.parse( (char *)ll );
        log->setAuditLevel( llevel.Value() );
    }

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::synchFS
// Description  : synchronize a filesystem
//
// Inputs       : fs - filesystem to synchronize
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::synchFS( AMFsys & fs )
{
    // Is this an import or export FS
    Boolean sent = FALSE;
    AMFsys *search = exports.find (&fs);
    if ( search != NULL )
    {
        // EXPORT : send update announcements
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "Update Export FS [%s]", (char *)fs );
        fs.scanEntInfo();
        fs.createAnnouncements( &updateAnnouncements );
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "End Update Expport FS [%s]", 
                 (char *)fs );
        sendUpdate();
    }
    else
    {
        // IMPORT : send update request
        sent = TRUE;
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "Update Import FS [%s]", (char *)fs );
        ((AMImpFsys *)&fs)->getUpdateRequests( &updateRequests, &random );
        sendUpdateRequest();
    }

    if ( sent == TRUE )
        LOGMSG1( log, AMIRD_ALEVEL, "Update Import FS [%s]", (char *)fs );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::xfersPending
// Description  : Are there file transfers pending?
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::xfersPending( AMFsysList *fsq )
{
    // Walk list of filesystems, looking for pending
    AMFsys * fs;
    AMFsysList::iterator it;
    for (it = fsq->begin(); it != fsq->end(); it++)
    {
        fs = *it;
        // If pending export.
        if ( fs->PendingUpdates().listSize() > 0 )
        {
            // Cleanup, return pending
            return( TRUE );
        }
    }

    // Return no exports pending
    return( FALSE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::spawnExportSubgroup
// Description  : spawn an export subgroup
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::spawnExportSubgroup( void )
{
    // Local variables
    AMFsys *fs;
    AMEntry * ent = NULL;

    // Walk list of exports, looking for exports pending
    AMFsysList::iterator iter;
    for (iter = exports.begin(); (ent == NULL) && (iter != exports.end()); iter++)
    {
        fs = *iter;
        // If pending export.
        if ( fs->PendingUpdates().listSize() > 0 )
        {
            Iterator *pu = fs->PendingUpdates().getIterator();
            AMEntry *e;
            while ((ent == NULL) &&
                (e = (AMEntry*)fs->PendingUpdates().next(pu)) != NULL)
            {
                if ((fs->InProgress().find(e)) == NULL)
                {
                    ent = e;
                }
            }
            delete pu;
        }
    }

    if (ent == NULL)
    {
        return FALSE;
    }

    // Now insert filesystem back onto export list (We get a round
    // robin of exports in this way)
    ATK_ASSERT( (fs != NULL), "Null filesysystem" );
    if ( (fs = exports.deleteItem(fs)) != NULL )
    {
        exports.push_back( fs );
    }

    // Any room to start a new subgroup and something to export?
    if ( (exportList.size() < (size_t)maxExportSgroups.Value()) &&
         (ent != NULL) )
    {
        // Scan for used ports
        unsigned short mport, sport, i;
        for ( i=1; (i<=maxExportSgroups.Value()) && 
                       (portsInUse.Value(i) == TRUE); i++ );
        if ( i <= maxExportSgroups.Value() )
        {
            // Pick the free ports
            mport = sgmcPort.Value() + i;
            sport = sgsrvrPort.Value() + i;
            portsInUse.Set( i, TRUE );
        }
        else
        {
            // Error out
            LOGMSG0( log, AMIRD_ERROR_LEVEL, "Exhausted export ports" );
            return( FALSE );
        }

        // Check the transport types, create transport
        ATransport *atrans = NULL;
        AString m = sgtransType.Value(), 
               mca = sgmcAddr.Value(), 
               sa = sgsrvrAddr.Value();
        atrans = initTransport( ATransport::transportType(m), isServer.Value(), 
                                mca, mport, sa, sport );

        // If bad transport, error out
        if ( atrans == NULL )
        {
            // Log the new subgroup
            LOGMSG0( log, AMIRD_REPORT_LEVEL, "Aborting subgroup, bad "
                     "transport" );
            portsInUse.Set( i, FALSE );
            return( FALSE );
        }

        // If no subgroup policy, default to control group policy
        AString & gpol = fs->Policy(),
               & lpol = fs->LocPolicy();

        // Now create the subgroup and add it to exported groups
        ATK_ASSERT( (ent->entryType() == AMEntry::ISFILE), "Bad export file" );
        AMFxfer * sgroup = new AMFxfer( atrans, (AMFile *)ent, gpol, lpol, log );

        // If errored, cleanup
        if ( sgroup->isErrored() )
        {
            // Log the new subgroup
            LOGMSG2( log, AMIRD_REPORT_LEVEL, "Aborting subgroup [%s,%s]",
                     (char *)*ent, (char *)atrans->toString() );

            // Cleanup group, return
            delete sgroup;
            delete atrans;
            portsInUse.Set( i, FALSE );
            return( FALSE );
        }
        else
        {
            // Mark begin, advertise the export group
            ent->xferBegin();
            exportList.push_back( sgroup );
            sendGroupAdversizement( sgroup );
    
            // Log the new subgroup
            LOGMSG2( log, AMIRD_REPORT_LEVEL, "Connected subgroup [%s,%s]",
                     (char *)sgroup->toString(), (char *)atrans->toString() );
        }
    }

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::shutdownExportSubgroup
// Description  : shutdown an export subgroup
//
// Inputs       : sgroup - the subgroup to shutdown
// Outputs      : none

void AMirD::shutdownExportSubgroup( AMFxfer *sgroup )
{
    // Cleanup the group
    ATransport *atrans = sgroup->Transport();
    if (atrans != NULL)
    {
        // Free the port for another group download
        int mport = sgroup->Transport()->MulticastAddress()->Port()-sgmcPort.Value();
        portsInUse.Set( mport, FALSE );

        // Set the file to completed (ignore failures, let request come again)
        sgroup->Entry()->xferComplete( TRUE );

        delete atrans;
    }
    delete sgroup;

    AMFxferList::iterator xfIter =
        find (exportList.begin(), exportList.end(), sgroup);

    if (*xfIter != NULL)
    {
        exportList.erase (xfIter);
    }
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::spawnImportSubgroup
// Description  : spawn an import subgroup
//
// Inputs       : ent - entry to import
//                md - subgroup multicast mode
//                addr - subgoup multicast address
//                pt - subgroup multicast port
//                saddr - subgroup server address
//                spt - subgroup server port
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::spawnImportSubgroup( AMEntry *ent, ATransport::AtModeEnum md, 
                                    unsigned long addr, unsigned short pt, 
                                    unsigned long saddr, unsigned short spt )
{
    // Do we have enough room in the import queue
    if ( importList.size() < (size_t)maxImportSgroups.Value() )
    {
        // Get the local policy
        AString & lpol = ent->Fsystem()->LocPolicy();

        // Create the transport, then the subgroup
        ATransport * atrans = 
        initTransport( md, isServer.Value(), addr, pt, saddr, spt );

        // Log the new subgroup
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "Connected subgroup [%s]",
                 (char *)atrans->toString() );

        // Create group, begin xfer
        AMFxfer * sgroup = new AMFxfer( atrans, (AMFile *)ent, lpol, log );
        importList.push_back( sgroup );
        ent->xferBegin();

        // Log the new subgroup
        LOGMSG2( log, AMIRD_REPORT_LEVEL, "Connected subgroup [%s,%s]",
                 (char *)sgroup->toString(), (char *)atrans->toString() );
    }

    // Return succesfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::shutdownImportSubgroup
// Description  : shutdown an import subgroup
//
// Inputs       : sgroup - the subgroup to shutdown
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::shutdownImportSubgroup( AMFxfer *sgroup )
{
    // Cleanup the group
    ATransport *atrans = sgroup->Transport();

    if (atrans != NULL)
    {
        // Set the file to completed
        sgroup->Entry()->xferComplete( (sgroup->Status() ==  AMFxfer::XFERCOMPL) );

        AMFxferList::iterator xfIter =
            find (importList.begin(), importList.end(), sgroup);
        if (*xfIter != NULL)
        {
            importList.erase (xfIter);
        }
        delete sgroup;
        delete atrans;
    }

    // If done with filesystem refresh LOG
    if ( xfersPending((AMFsysList*)(&imports)) == FALSE )
    {
        // Start the download clock
        LOGMSG1( log, AMIRD_ALEVEL, "END FILESYSTEM UPDATE (%lu secs)",
                 AtkClock::SysClock()-downloadTime );
    }

    // Return successfully
    return( TRUE );
}

//
// Message Routines

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendInterface
// Description  : send a message to the group
//
// Inputs       : buf - buffer to send
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendInterface( Buffer * buf )
{
#ifdef DROP_TEST
#define DROP_RATE 5 // Percent
    // Drop a percentage of the packets
    if ( Random::GenerateRandom(0, 99) < DROP_RATE )
    {
        // Cleanup buffer, return successfully
        delete buf;
        return( TRUE );
    }
#endif

    // Send the message
    LOGMSG1( log, AMIRD_DETAIL_LEVEL, "Send [%s]", 
             messageString((MessageTypes)ntohs(*(unsigned short *)buf->Buf())) );
    ctrlGroup->sendMessage( buf );
    if ( ctrlGroup->State() != AGroup::AG_INOPERATION )
    {
        // Error out
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "[%s] send failed", 
                 messageString(*(MessageTypes *)buf->Buf()) );
        shutdown = TRUE;
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::recvInterface
// Description  : send a message to the group
//
// Inputs       : buf - buffer to send
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::recvInterface( Buffer **buf )
{
    // Read the message, check for error state
    struct timeval tm = { 0, 0 };
    ctrlGroup->readMessage( buf, &tm );
    if ( ctrlGroup->State() != AGroup::AG_INOPERATION )
    {
        // Error out
        LOGMSG0( log, AMIRD_ERROR_LEVEL, "amird recv failed" );
        shutdown = TRUE;
        return( FALSE );
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processMessage
// Description  : receive and process incoming message
//
// Inputs       : void
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::processMessage( void )
{
    // Try once, repeat while more data to be read
    do
    {
        // Read the message, check for error state
        Buffer * buf = NULL;
        if ( recvInterface(&buf) == FALSE ) return( FALSE );
    
        // If nothing received, just return 
        if ( buf == NULL ) return( FALSE );
    
        // Get type, log the reception
        unsigned short mt;
        *buf >> mt;
        LOGMSG1( log, AMIRD_DETAIL_LEVEL, "Recv [%s]", 
                 messageString((MessageTypes)mt) );
    
        // Switch on message type
        switch ( mt )
        {
            case AMIRD_FS_UPDATE: // Filesystem update
                 processUpdate( buf );
                 break;

            case AMIRD_FS_SYNCREQ: // Filesystem update request
                 processSynchReq( buf );
                 break;
    
            case AMIRD_FS_REQUEST: // Request for update
                 processUpdateRequest( buf );
                 break;
    
            case AMIRD_FS_SUBGRP: // Update group announcement
                 processGroupAdversizement( buf );
                 break;
    
            case AMIRD_FS_DELETE: // Client telling us that a file it has should be deleted.
                 processClientFileDelete( buf );
                 break;
    
            case AMIRD_FS_QUIT: // quit
                 LOGMSG0(log, AMIRD_REPORT_LEVEL,
                     "AMirD terminating normally due to server shutdown.");
                 shutdown = TRUE;
                 break;
    
            case AMIRD_FS_UNKNOWN: // Unknown message
            default: // Unknown message type, error out
                 LOGMSG1( log, AMIRD_ERROR_LEVEL, "Unknown message type (%d)", mt );
                 shutdown = TRUE;
                 return( FALSE );
        }
    
        // Cleanup the message
        delete buf;
    }
    while ( ctrlGroup->messagePending() == TRUE );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendUpdate
// Description  : send an update for a filesystem
//
// Inputs       : fs - filesystem to announce update
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendUpdate( void )
{
    // Get the announcement
    static Buffer abuf;
    Buffer *annce;

    // Check existance of update requests
    if (!updateAnnouncements.empty())
    {
        annce = updateAnnouncements.front();
        updateAnnouncements.pop_front();
        // Copy contents, construct message
        abuf = *annce;
        annce->Reset();
        LOGMSG2(log, AMIRD_DETAIL_LEVEL,
            "Config %s : %d",
            (char*)maxExportSgroups.Name(), maxExportSgroups.Value());

        LOGMSG2(log, AMIRD_DETAIL_LEVEL,
            "Config %s : %d",
            (char*)maxImportSgroups.Name(), maxImportSgroups.Value());

        (*annce) << (unsigned short)AMIRD_FS_UPDATE << globalCheckOnly << abuf;
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "Server Mode: %s",
            globalCheckOnly ? "Check files without changing." : "Modify files");
        // (*annce) << (unsigned short)AMIRD_FS_UPDATE << abuf;
        // Send the announcement
        if ( sendInterface( annce ) == FALSE )
        {
            // Error out
            LOGMSG0( log, AMIRD_ERROR_LEVEL, "Update send failed" );
            return( FALSE );
        }
    
        // Reset, if needed, annoucement timer
        if ( updateAnnouncements.size() > 0 )
        {
            updateAnnounceTimer.reset();
            exQueue.addTimer( &updateAnnounceTimer );
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processUpdate
// Description  : process an incoming update
//
// Inputs       : msg - message containing FS details
// Outputs      : TRUE if successful, FALSE otherwise

void AMirD::processUpdate( Buffer * msg )
{
    LOGMSG2(log, AMIRD_DETAIL_LEVEL,
        "Config %s : %d",
        (char*)maxExportSgroups.Name(), maxExportSgroups.Value());

    LOGMSG2(log, AMIRD_DETAIL_LEVEL,
        "Config %s : %d",
        (char*)maxImportSgroups.Name(), maxImportSgroups.Value());

    // Get the checkOnly flag.  This ensures that the
    // server and client are both in the same mode.
    // Note that if any exporter ever sets this flag,
    // then it will propagate to other importers, which
    // may also be exporters, which will in turn
    // propagate it to other importers.
    *msg >> globalCheckOnly;
    LOGMSG1( log, AMIRD_REPORT_LEVEL, "Client Mode: %s",
        globalCheckOnly ? "Check files without changing." : "Modify files");

    // Extract the announcement
    Buffer annce;
    *msg >> annce;

    // Is this the beginning of a new set of downloads?
    Boolean start = (xfersPending( (AMFsysList*)(&imports) ) == FALSE);

    // Walk the imported filesystem to see if we care
    AMImpFsys *fs;

    AMImpFsysList::iterator fsImpIter;
    for (fsImpIter = imports.begin(); fsImpIter != imports.end(); fsImpIter++)
    {
        fs = *fsImpIter;
        // Check to see if this FS cares
        if ( fs->processAnnouncement(annce) == TRUE )
        {
            // Sync/send update requests
            synchFS( *fs );

            // If this is beginning of new downloads
            if ( (start == TRUE) && (fs->PendingUpdates().listSize() > 0) )
            {
                // Start the download clock
                LOGMSG0( log, AMIRD_ALEVEL, "Start FILESYSTEM UPDATE" );
                downloadTime = AtkClock::SysClock();
            }
        }
    }

    // Log unprocessed update, cleanup
    LOGMSG0( log, AMIRD_DETAIL_LEVEL, "Update unproced" );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendSynchReq
// Description  : send an request for a filesystem update
//
// Inputs       : fs - filessytem to request
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendSynchReq( AMImpFsys *fs )
{
    // Create the sync-req
    Buffer *qbuf = new Buffer();
    (*qbuf) << (unsigned short)AMIRD_FS_SYNCREQ
            << fs->Exporter() << fs->ExportPath();

    // Send the announcement
    if ( sendInterface( qbuf ) == FALSE )
    {
        // Error out
        LOGMSG0( log, AMIRD_ERROR_LEVEL, "Update request send failed" );
        return( FALSE );
    }

    // Return successfully    
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processSynchReq
// Description  : Process an incoming synch request
//
// Inputs       : fs - filessytem to request
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::processSynchReq( Buffer *msg )
{
    // Get the exporter and filesystem
    AMFsys *fs;
    AString ex, fsys;
    (*msg) >> ex >> fsys;

    // If meant for me and good filesystem
    if ( (ex == exporter.Value()) &&
         ((fs = (AMFsys *)exports.find( &fsys )) != NULL) )
    {
        // If allowed, do sync
        if ( fs->allowUpdate() == TRUE )
        {
            // Synch, Log synch req
            synchFS( *fs );
            LOGMSG2( log, AMIRD_REPORT_LEVEL, "SYNCH-REQ Processed [%s/%s]", 
                     (char *)ex, (char *)fsys );
        }
        else
        {
            // Log ignoring
            LOGMSG2( log, AMIRD_REPORT_LEVEL, "SYNCH-REQ Ignored (min update)"
                     " [%s/%s]", (char *)ex, (char *)fsys );
        }
    }
    else
    {
        // Log ignoring
        LOGMSG2( log, AMIRD_DETAIL_LEVEL, "SYNCH-REQ Ignored [%s/%s]",
                 (char *)ex, (char *)fsys );
    }

    // Return succesfully
    return( TRUE );
}


//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendUpdateRequest
// Description  : send an update request to an exporter
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendUpdateRequest( void )
{
    // Get the announcement
    static Buffer abuf;
    Buffer *annce;

    // Check existance of update requests
    if (!updateRequests.empty())
    {
        annce = updateRequests.front();
        updateRequests.pop_front();
        // Copy contents, construct message
        abuf = *annce;
        annce->Reset();
        *annce << (unsigned short)AMIRD_FS_REQUEST << abuf;
    
        // Send the announcement
        if ( sendInterface( annce ) == FALSE )
        {
            // Error out
            LOGMSG0( log, AMIRD_ERROR_LEVEL, "Update request send failed" );
            return( FALSE );
        }
    
        // Reset, if needed, annoucement timer
        if (!updateRequests.empty())
        {
            updateRequestTimer.reset();
            exQueue.addTimer( &updateRequestTimer );
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processUpdateRequest
// Description  : process incoming update request
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::processUpdateRequest( Buffer * msg )
{
    // Extract the announcement
    Buffer req;
    *msg >> req;

    // First get the exporter and filesystem
    AString ex, fss;
    if ( (AMFsys::getFsysFromAnnounce(req,ex,fss) == TRUE) &&
         (ex == exporter.Value()) )
    {
        // Look for the export/import filesystem to see if we care
        AMFsys *fs = exports.find(&fss);
        if ( fs != NULL )
            // Look in exports
            fs->processUpdateRequest( &req );
        else
        {
            // Ok, look in imports
            if ( (fs = imports.find(&fss)) != NULL )
                fs->processUpdateRequest( &req );
        }
    }

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendGroupAdversizement
// Description  : send a group advertisement
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendGroupAdversizement( AMFxfer * sgroup )
{
    // Create the buffer put message header information
    AString enm;
    Buffer *buf = new Buffer(), *b2;
    b2 = sgroup->Entry()->Fsystem()->createFileId(sgroup->Entry());
    *buf << (unsigned short)AMIRD_FS_SUBGRP
         << *b2
         << (unsigned short)sgroup->Transport()->Mode()
         << sgroup->Transport()->MulticastAddress()->Address()
         << sgroup->Transport()->MulticastAddress()->Port();
    delete b2;

    // If not doing symmetric multicast
    if ( sgroup->Transport()->Mode() != ATransport::AT_SYMMETRIC )
    {
        // add server address and port
        *buf << sgroup->Transport()->ServerAddress()->Address()
             << sgroup->Transport()->ServerAddress()->Port();
    }

    // Send the message to the control group
    if ( sendInterface( buf ) == FALSE )
    {
        // Error out
        LOGMSG0( log, AMIRD_ERROR_LEVEL, "Update send failed" );
        return( FALSE );
    }

    // Log update send
    LOGMSG1( log, AMIRD_REPORT_LEVEL, "Sent group advertisement [%s]", 
             (char *)sgroup->toString() );

    // Return successfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processGroupAdversizement
// Description  : process a group advertisement
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::processGroupAdversizement( Buffer * msg )
{
    // If not importing anything, bail
    if ( imports.size() == 0 ) return( FALSE );

    // Local vars
    unsigned short md, pt, spt = 0;
    unsigned long addr, saddr = 0;
    Buffer id;
    AMEntry *ent = NULL;
    AMFsys *fs = NULL;

    // Read the entry, addressing and mode information
    *msg >> id >> md >> addr >> pt;

    // Find the file, filesystem associated with this announcement
    AMImpFsysList::iterator fsImpIter;
    for (fsImpIter = imports.begin();
         (ent == NULL) && (fsImpIter != imports.end());
         fsImpIter++)
    {
        fs = *fsImpIter;
        ent = fs->findFileById(&id);
    }

    // If not found or not pending, bail
    if ( (ent == NULL) || (ent->PendingUpdate() == FALSE) || 
         (ent->Fsystem()->InProgress().find(ent) != NULL) )
    {
        LOGMSG1( log, AMIRD_REPORT_LEVEL, "Ignoring announcement for [%s]", 
                 (ent == NULL) ? "unknown" : (char *)*ent );
        return( FALSE );
    }

    // Check if the mode requires more information
    if ( md != ATransport::AT_SYMMETRIC )
        *msg >> saddr >> spt;

    // Spawn the import group, log
    spawnImportSubgroup( ent, (ATransport::AtModeEnum)md, addr, pt, saddr, spt );
    LOGMSG1( log, AMIRD_REPORT_LEVEL, "Recv SG announment [%s]", (char *)*ent );

    // Return successfully
    return( TRUE );
}

//
// Static Private Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::initTransport
// Description  : initialize a transport group 
//
// Inputs       : mode - transport mode
//                iss - server flag
//                mca - multicast address
//                mcp - multicast port
//                sa - server address
//                sp - server port
// Outputs      : TRUE if successful, FALSE otherwise

ATransport * AMirD::initTransport( ATransport::AtMode mode, Boolean iss,
                                   AString & mca, unsigned short mcp, 
                                   AString & sa, unsigned short sp )
{
    // If going to use multicast, create IPADDR
    IPAddress *mipa = NULL, *sipa = NULL;
    if ( mode != ATransport::AT_PTTOPT )
    {
        // Setup the multicast address
        mipa = IPAddress::IPAddressFactory( mca, mcp );
        if ( mipa == NULL )
        {
            // Error out
            LOGMSG2( log, AMIRD_ERROR_LEVEL, "Bad multicast address [%s/%d]", 
                     (char *)mca, mcp );
            return( FALSE );
        }
    }

    // If going to use server addr, create IPADDR
    if ( (mode != ATransport::AT_SYMMETRIC) || 
         ((mode == ATransport::AT_ASYMMETRIC ) && (iss == FALSE)) )
    {
        // Setup the server address
        sipa = IPAddress::IPAddressFactory( sa, sp );
        if ( sipa == NULL )
        {
            // Error out
            LOGMSG2( log, AMIRD_ERROR_LEVEL, "Bad server address [%s/%d]", 
                     (char *)sa, sp );
            return( FALSE );
        }
    }

    // Create the transport layer
    ATransport *trans;
    if ( iss == TRUE )
        trans = new ATransport( mipa, (unsigned short)sp, mode );
    else
        trans = new ATransport( mipa, sipa, mode );
    trans->Log() = log;
    trans->SetLoopback( lookbackMCast.Value() );

    // Connected
    if ( trans->Connect() == FALSE )
    {
        // Error out
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Error creating transport [%s]", 
                 (char *)trans->toString() );
        return( FALSE );
    }

    // Return succesfully
    return( trans );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::initTransport
// Description  : initialize a transport group 
//
// Inputs       : mode - transport mode
//                iss - server flag
//                mca - multicast address
//                mcp - multicast port
//                sa - server address
//                sp - server port
// Outputs      : TRUE if successful, FALSE otherwise

ATransport * AMirD::initTransport( ATransport::AtMode mode, Boolean iss,
                                   unsigned long mca, unsigned short mcp, 
                                   unsigned long  sa, unsigned short sp )
{
    // If going to use multicast, create IPADDR
    IPAddress *mipa = NULL, *sipa = NULL;
    if ( mode != ATransport::AT_PTTOPT )
    {
        // Setup the multicast address
        mipa = IPAddress::IPAddressFactory( mca, mcp );
        if ( mipa == NULL )
        {
            // Error out
            LOGMSG2( log, AMIRD_ERROR_LEVEL, "Bad multicast address [%s/%d]", 
                     (char *)mca, mcp );
            return( NULL );
        }
    }

    // If going to use server addr, create IPADDR
    if ( (mode != ATransport::AT_SYMMETRIC) || 
         ((mode == ATransport::AT_ASYMMETRIC ) && (iss == FALSE)) )
    {
        // Setup the server address
        sipa = IPAddress::IPAddressFactory( sa, sp );
        if ( sipa == NULL )
        {
            // Error out
            LOGMSG2( log, AMIRD_ERROR_LEVEL, "Bad server address [%s/%d]", 
                     (char *)sa, sp );
            return( NULL );
        }
    }

    // Create the transport layer
    ATransport *trans;
    if ( iss == TRUE )
        trans = new ATransport( mipa, (unsigned short)sp );
    else
        trans = new ATransport( mipa, sipa );
    trans->Log() = log;
    trans->SetLoopback( lookbackMCast.Value() );

    // Connected
    if ( trans->Connect() == FALSE )
    {
        // Error out
        LOGMSG1( log, AMIRD_ERROR_LEVEL, "Error creating transport [%s]", 
                 (char *)trans->toString() );
        return( NULL );
    }

    // Return succesfully
    return( trans );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::sendQuit
// Description  : send a message to quit to the whole group.
//                This is the normal, orderly way to shutdown.
//
// Inputs       : none
// Outputs      : TRUE if successful, FALSE otherwise

Boolean AMirD::sendQuit (void)
{
    // Create the sync-req
    Buffer *buf = new Buffer();
    (*buf) << (unsigned short)AMIRD_FS_QUIT;

    // Send the announcement
    if (!sendInterface(buf))
    {
        // Error out
        LOGMSG0( log, AMIRD_ERROR_LEVEL, "Attempt to broadcast QUIT failed." );
        return FALSE ;
    }

    // Return successfully    
    return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirD::processClientFileDelete
// Description  : When in checkOnly mode, clients tell the server
//                which files they have that should be
//                deleted (deleted at the client).
//
// Inputs       : msg - message containing FS details
// Outputs      : TRUE if successful, FALSE otherwise

void AMirD::processClientFileDelete( Buffer * msg )
{
    // Extract the client name and file name

    AString clientName;
    AString fileName;

    *msg >> clientName >> fileName;

    LOGMSG2 (log, AMIRD_REPORT_LEVEL, "Client %s needs to delete file %s.",
        (char*)clientName, (char*)fileName);

    LOGMSG2 (report, AMIRD_REPORT_LEVEL, "%s delete %s",
        (char*)clientName, (char*)fileName);

}
