#ifndef AMirD_INCLUDED
#define AMirD_INCLUDED

/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
*
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
*
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMirD.h
//
//   Created : 25-Sep-00
//   By : Patrick McDaniel
//
//   Description : This is the base class for the AMirD application.
//                 It mirrors a set of directories specified in a the
//                 configuration file.
//
//   *** Change History ***
//
//   25-Sep-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// *** Include Files ***
#include <AGroup.h>
#include <StringVar.h>
#include <IntVar.h>
#include <BooleanVar.h>
#include <Environment.h>
#include <Certificate.h>

// Project Include Files
#include <AMImpFsys.h>
#include <AMFxfer.h>

// *** Definitions ***
#define CONFIG_SECTION_TEXT "[config]"
#define EXPORTS_SECTION_TEXT "[exports]"
#define IMPORTS_SECTION_TEXT "[imports]"

#define GRPPOL_CONFIG "grppol"
#define LOCPOL_CONFIG "locpol"
#define EXPORTER_CONFIG "exporter"
#define UPPERIOD_CONFIG "updateperiod"
#define NUMPERIOD_CONFIG "numperiod"
#define CHECKONLY_CONFIG "checkonly"
#define ISSRVR_CONFIG "isserver"
#define MAX_ESG_CONFIG "maxexportsubgroups"
#define MAX_ISG_CONFIG "maximportsubgroups"
#define AUTHOR_CERT "authoritycert"
#define TRANSPT_CONFIG "transport"
#define MCADDR_CONFIG "mcaddr"
#define MCPORT_CONFIG "mcport"
#define SRVADDR_CONFIG "srvraddr"
#define SRVPORT_CONFIG "srvrport"
#define SGTRANSPT_CONFIG "sgtransport"
#define SGADDR_CONFIG "sgaddr"
#define SGPORT_CONFIG "sgport"
#define SGSRVAD_CONFIG "sgsrvraddr"
#define SGSRVPT_CONFIG "sgsrvrport"
#define SYMLINKS_CONFIG "followsymlinks"
#define LOOPBACK_CONFIG "mcloopback"
#define MINUPDATE_CONFIG "minupdateperiod"

#define DEFAULT_UPDATE_PERIOD 10 // (seconds)
#define DEFAULT_NUM_PERIOD -1 // (number of periods.  -1: infinite)
#define DEFAULT_CHECK_ONLY FALSE // (if TRUE, only check files, do not transfer)
#define INTER_ANNOUNCE_PERIOD 10 // (milliseconds)
#define INTER_REQUEST_PERIOD 10 // (milliseconds)
#define MAX_EXPORT_SUBGRPS "1"
#define MAX_IMPORT_SUBGRPS "1"
#define HARD_MAX_SUBGROUPS 64  // ports in use
#define DEF_SYMLINK_ENABLE "TRUE"
#define DEF_MIN_UP_PERIOD "30"

// Forward References
class AMirDPolicyImpl;

// Class Definition AMirD
class AMirD {

  public: // Type Definitions

  // The types of configuration states (reading config file)
  typedef enum {
      AMIRD_CFG_UNKNOWN = 0,
      AMIRD_CFG_VARS    = 1,
      AMIRD_CFG_EXPORTS = 2,
      AMIRD_CFG_IMPORTS = 3
  } configState;

  // The message types
  typedef enum {
      AMIRD_FS_UNKNOWN = 0, // Unknown message
      AMIRD_FS_UPDATE  = 1, // Filesystem update
      AMIRD_FS_SYNCREQ = 2, // Filesystem update request
      AMIRD_FS_REQUEST = 3, // Request for update
      AMIRD_FS_SUBGRP  = 4, // Update group announcement
      AMIRD_FS_DELETE  = 5, // Deleted a file
      AMIRD_FS_QUIT    = 6, // Group should stop
      AMIRD_FS_MAXMSG  = 7  // Maximum message number
  } MessageTypes;

  // The timer types
  typedef enum {
      AMIRD_UNKNOWN       = 0, // Unknown message
      AMIRD_UPDATE_TIMER  = 1, // Send update announce/requests
      AMIRD_EXPORT_SND    = 2, // Space export announcements
      AMIRD_IMPORT_REQ    = 3, // Space import requests
  } TimerTypes;

  public: // Public Interface

    //
    // Constructors and Destructors

    AMirD( AString & fname, AString * logName = NULL );
      // Default Constructor

    virtual ~AMirD( void );
      // Destructor

    //
    // Access Methods

    //
    // Action Methods

    // Re-read the configurations.
    void reconfigure( void )
    {
        reconfig = TRUE;
    }

    // Initiate a synch
    void resync( void )
    {
        synchronize = TRUE;
    }

    // shutdown the application
    void shutDown( void )
    {
        shutdown = TRUE;
    }

    // Get the log associated with this application
    AuditLog * Log( void )
    {
        return( log );
    }

    // list of exports
    AMFsysList *Exports( void )
    {
        return &exports;
    }

    // list of imports
    AMImpFsysList *Imports( void )
    {
        return &imports;
    }

    // This is a reference to the Antigone Group Object
    AGroup * ControlGroup( void )
    {
        return( ctrlGroup );
    }

    //
    // Member Methods

    Boolean execute( void );
      // execute the application loop

    void LogConfiguration( unsigned long lvl );
      // Log the current configuration.

    //
    // Static Member Methods

    // Return a string describing the message
    static const char * messageString( MessageTypes m )
    {
        if ( (m < AMIRD_FS_UNKNOWN) || (m >= AMIRD_FS_MAXMSG) )
            return( "Unknown" );
        return( messageStrings[m] );
    }

    // Return flag indicating whether symbolic links should be followed
    static Boolean FollowSymbolicLinks( void )
    {
        return( followSymLinks.Value() );
    }

    // Get the resident authority
    static AString & GetAuthority( AString & out )
    {
        // If authority defined
        Environment::getStringEnvVar( "AUTHOR", out );
        return( out );
    }

    static Boolean getAttributeParm( AGroup *grp, Var & var,
                                     const char *def, AuditLog *lg );
      // Get get an attribute by policy

    Boolean sendInterface( Buffer * buf );
      // send a message to the group

    // get the report log
    AuditLog *Report(void)
    {
        return report;
    }

  private: // Private Interface and Data

    //
    // Private Interfaces

    Boolean processTimers( AuditLog *lg );
      // process all expired timers

    Boolean readConfigFile( void );
      // Read the amird configuration file

    Boolean parseConfigFileLine( configState &cfg, unsigned char * ln );
      // parse one line of the configuration

    Boolean parseConfigLine( AString & line );
      // parse a string containing a configuration value

    Boolean parseExportLine( AString & line, AString & fs,
                             AString & gp, AString & lp, AString & at );
      // parse a string containing a export line

    Boolean parseImportLine( AString & line, AString & expt, AString & fs,
                             AString & mpt, AString & lp, AString & at );
      // parse a string containing a import line

    void cleanup( void );
      // cleanup groups and configuration

    AMFsys * findImport( AMFsys & fs );
      // find an import associated with the export update

    Boolean initializeControlGroup( void );
      // initialize the control group

    Boolean synchFS( AMFsys & fs );
      // synchronize a filesystem

    Boolean xfersPending( AMFsysList *fsq );
      // Are there file transfers pending?

    Boolean spawnExportSubgroup( void );
      // spawn a export subgroup

    void shutdownExportSubgroup( AMFxfer *sgroup );
      // shutdown an export subgroup
   
    Boolean spawnImportSubgroup( AMEntry *ent, ATransport::AtModeEnum md,
                                 unsigned long addr, unsigned short pt,
                                 unsigned long saddr, unsigned short spt );
      // Spawn a export subgroup

    Boolean shutdownImportSubgroup( AMFxfer *sgroup );
      // shutdown an import subgroup

    void sendCheckRequests ( void );

    //
    // Message Routines

    Boolean recvInterface( Buffer **buf );
      // send a message to the group

    Boolean processMessage( void );
       // receive and process incoming message

    Boolean sendUpdate( void );
      // send an update for a filesystem

    void processUpdate( Buffer * msg );
      // process an incoming update

    Boolean sendSynchReq( AMImpFsys *fs );
      // send an request for a filesystem update

    Boolean processSynchReq( Buffer *msg );
      // Process an incoming synch request

    Boolean sendUpdateRequest( void );
      // send an update request to an exporter

    Boolean processUpdateRequest( Buffer * buf );
      // process incoming update request

    Boolean sendGroupAdversizement( AMFxfer * sgroup );
      // send a group advertisement

    Boolean processGroupAdversizement( Buffer * msg );
      // process a group advertisement

    void processClientFileDelete( Buffer * msg );
      // process the notification from a client that a
      // file needs to be deleted (at the client)

    //
    // Static Private Methods

    ATransport * initTransport( ATransport::AtMode mode, Boolean iss,
                                AString & mca, unsigned short mcp,
                                AString & sa, unsigned short sp );
      // initialize a transport group

    ATransport * initTransport( ATransport::AtMode mode, Boolean iss,
                                unsigned long mca, unsigned short mcp,
                                unsigned long  sa, unsigned short sp );

    Boolean sendQuit (void);
      // tell the group to quit.

    //
    // Class Attributes

    AuditLog *log;
      // This is the amird log.

    AString configFile;
      // This is the configuration file for this application.

    AMFsysList exports;
      // List of directories to mirror

    AMImpFsysList imports;
      // List of directories to receive mirroring

    AtkTimerQueue exQueue;
      // Queue of export timers.

    Boolean shutdown;
      // Shutdown the application?

    Boolean reconfig;
      // Reconfigure the application?

    Boolean synchronize;
      // Flag indicating filesystem should be synchronized

    unsigned long  downloadTime;
      // This is the total time it takes to resynch the filesystem.

    //
    // Update Announcement/Request Stuff

    AtkTimer updateAnnounceTimer;
      // This timer is used to spread out the updates.

    BufferList updateAnnouncements;
      // This is the list of announcements for exports.

    AtkTimer updateRequestTimer;
      // This timer is used to spread out the requests.

    BufferList updateRequests;
      // This is the list of request for imports.

    //
    // Group Object configurationn

    ATransport *ctrlTrans;
      // This is the transport object for the control group.

    AGroup *ctrlGroup;
      // This is the control group

    Certificate *authCert;
      // This is the control group authority certificate

    AAttributeSet attrs;
      // This is the attribute set for the group.

    AAttribute authAttr;
      // This is the authority attribute for the group

    AAttribute authIdAttr;
      // This is the authority id attribute for the group

    AMirDPolicyImpl *polImpl;
      // This the control group policy implementor

    AMFxferList exportList;
      // list of export subgroups

    AMFxferList importList;
      // list of import subgroups

    BitVector portsInUse;
      // This is a list of the ports in use.

    //
    // Configuration Values

    StringVar grpPolicy, locPolicy;
      // These are the filename of the group and local policies.

    StringVar exporter;
      // This is the name under which the filesystems are exported.

    IntVar upPeriod;
      // This is the periodicity of updates

    IntVar numPeriod;
      // quit after this number of update periods. -1 means
      // perform updates forever.

    BooleanVar checkOnly;
      // if TRUE, only check for file updates that
      // need to be done, but do not actually transfer
      // the files.

    BooleanVar isServer;
      // This is a flag indicating the process is the group server.

    IntVar maxExportSgroups;
      // This is the maximum number of export subgroups.

    IntVar maxImportSgroups;
      // This is the maximum number of import subgroups.

    StringVar authorityCert;
      // This is the certificate used as the authority

    StringVar transType;
      // This is the transport type used for mcast.

    StringVar mcAddr;
      // This is the multicast address used for the application

    IntVar mcPort;
      // This is the multicast port used for the application

    StringVar srvrAddr;
      // This is the multicast address used for the application

    IntVar srvrPort;
      // This is the multicast port used for the application

    StringVar sgtransType;
      // This is the transport type used for mcast.

    StringVar sgmcAddr;
      // This is the multicast address used to export files (should be
      // unique to each exporter)

    IntVar sgmcPort;
      // This is the multicast port used for the application

    StringVar sgsrvrAddr;
      // This is the multicast address used for the application

    IntVar sgsrvrPort;
      // This is the multicast port used for the application

    // base for random number generation
    Random random;

    AuditLog *report;
      // send reporting messages to this log

    //
    // Static Class Attributes

    static const char * messageStrings[AMIRD_FS_MAXMSG];
      // These are description strings for the messages.

    //
    // Static Configuration Values

    static BooleanVar followSymLinks;
      // This is a flag indicating symbolic links should be followed.

    static IntVar minUpdatePeriod;
       // This is the minimum update period

    static BooleanVar lookbackMCast;
      // This is a flag multicast loopback should be used.

};
// End Class Definition AMirD

#endif

