/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : AMirDPolicyImpl.cpp
//
//   Created : 22-May-01
//   By : Patrick McDaniel
//
//   Description : (See AMirDPolicyImpl.h for class description)
//
//   *** Change History ***
//
//   22-May-01 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <pwd.h>
#include <sys/types.h>

// Project Include Files
#include <AMirDPolicyImpl.h>
#include <AMirD.h>
#include <AMFxfer.h>
#include <Environment.h>

// Definitions

//
// Static Class Attributes
const char * AMirDPolicyImpl::AMIRD_SENSITIVE_DIRECTORIES[AM_MAX_SENS_DIRS] = { 
    "/etc", 
    "/home/pdmcdan/thesis"
  };

const char * AMirDPolicyImpl::AMIRD_SENSITIVE_SITES[AM_MAX_SENS_SITES] = { 
    "initiator", 
  };

// 
// Member Methods

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirDPolicyImpl::EvaluateCondition
// Description  : Evaluate an application conditional
//
// Inputs       : cond - conditional structure
// Outputs      : TRUE if evaluated, FALSE otherwise

Boolean AMirDPolicyImpl::EvaluateCondition( ACondEval & cond )
{
    // Get the conditional predicate
    AString pred = APolicyImplementor::Predicate( cond );
    if ( strcasecmp((char *)pred, "iscontrolgroup") == 0 )
    {
        // Control group? - Test, set evaluation result information
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = (groupType == AMIRD_CONTROL_GROUP) ? TRUE : FALSE;
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "isdownloadgroup") == 0 )
    {
        // Download group? - Test, set evaluation result information
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = (groupType == AMIRD_CONTROL_GROUP) ? FALSE : TRUE;
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "injoinphase") == 0 )
    {
        // Not implemented for control groups (really non-sensical)
        if ( groupType == AMIRD_CONTROL_GROUP ) return( FALSE );

        if ( (dwnldGroup->ProtocolPhase() == AMFxfer::START) ||
             (dwnldGroup->ProtocolPhase() == AMFxfer::WAIT_START) )
        {
            // Always eval if in join phase 
            cond.ResultType() = ACondEval::APDL_ALWAYS_EVAL;
            cond.Result() = TRUE;
            LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate ALWAYS EVAL %s", (char *)pred,
                     BOOLE_STR(cond.Result()) );
        }
        else
        {
            // Once out of join phase, will never get back -> INVARIENT
            cond.ResultType() = ACondEval::APDL_INVARIANT;
            cond.Result() = FALSE;
            LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred,
                     BOOLE_STR(cond.Result()) );
        }
    }
    else if ( strcasecmp((char *)pred, "hasReadAccess") == 0 )
    {
        // Not implemented for control groups (really non-sensical)
        if ( groupType == AMIRD_CONTROL_GROUP ) return( FALSE );

        // Check argumements, get id and file
        AString id, file;
        if ( (APolicyImplementor::NoArgs(cond) != 2) || 
             (APolicyImplementor::Arg(0,cond,id) == FALSE) ||
             (APolicyImplementor::Arg(1,cond,file) == FALSE) )
        {
            // Error out
            LOGMSG1( log, AGRP_ALEVEL, "Malformed hasReadAccess condition [%s]",
                     (char *)cond );
            return( FALSE );
        }

        // Get the user/file information
        struct passwd *uinfo;
        struct stat finfo;
        // id = "pdmcdan";
        if ( ((uinfo = getpwnam((char *)id)) == NULL) ||
             (stat((char *)file,&finfo) != 0) )
        {
            // Log the error, return
            if ( uinfo == NULL )
            {
                // Log the account failure
                LOGMSG1( log, AGRP_ALEVEL, "Unable to map entity to UNIX "
                         "account [%s]", (char *)cond );
            }
            else
            {
                // Log the file failure
                LOGMSG1( log, AGRP_ALEVEL, "Unable to find file [%s]", 
                         (char *)cond );
            }
            return( FALSE );
        }

        // OK check to see if access allowed
        cond.ResultType() = ACondEval::APDL_ALWAYS_EVAL;
        cond.Result() = ((finfo.st_mode&S_IRUSR) &&  // User
                         (uinfo->pw_uid == finfo.st_uid)) ||
                        ((finfo.st_mode& S_IRGRP) && // Group
                         (uinfo->pw_gid == finfo.st_gid)) ||
                        (finfo.st_mode&S_IROTH);     // World
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate ALWAYS_EVAL %s", 
                 (char *)cond, BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "hassensitivefilesystem") == 0 )
    {
        // Not implemented for download groups (really non-sensical)
        if ( groupType == AMIRD_DOWNLOAD_GROUP ) return( FALSE );

        // Check export groups
        AMFsys *fsys;
        AString lpath;
        Boolean res = FALSE;
        AMFsysList *exports = cntrlGroup->Exports();
        AMFsysList::iterator exIt;
        
        for (exIt = exports->begin(); (!res) && (exIt != exports->end()); exIt++)
        {
            fsys = *exIt;
            res = isSensitive(fsys->localPath(lpath));
        }

        // If still needed
        if ( res == FALSE )
        {
            // Check import groups
            AMImpFsysList *imports = cntrlGroup->Imports();
            AMImpFsysList::iterator imIt;

            for (imIt = imports->begin(); (!res) && (imIt != imports->end()); imIt++)
            {
                fsys = *imIt;
                res = isSensitive(fsys->localPath(lpath));
            }
        }

        // Test whether the file is sensitive
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = res;
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "issensitivefile") == 0 )
    {
        // Not implemented for control groups (really non-sensical)
        if ( groupType == AMIRD_CONTROL_GROUP ) return( FALSE );

        // Check argumements, get file
        AString file;
        if ( (APolicyImplementor::NoArgs(cond) != 1) || 
             (APolicyImplementor::Arg(0,cond,file) == FALSE) )
        {
            // Error out
            LOGMSG1( log, AGRP_ALEVEL, "Malformed isSensitiveFile "
                     "condition [%s]", (char *)cond );
            return( FALSE );
        }

        // Test whether the file is sensitive
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = isSensitive(file);
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "issensitivesite") == 0 )
    {
        // Check argumements, get test group type
        AString site;
        if ( (APolicyImplementor::NoArgs(cond) != 1) || 
             (APolicyImplementor::Arg(0,cond,site) == FALSE) )
        {
            // Error out
            LOGMSG1( log, AGRP_ALEVEL, "Malformed groupType "
                     "condition [%s]", (char *)cond );
            return( FALSE );
        }

        // Is this a sensitive site?
        Boolean sens = FALSE;
        for ( int i=0; i<AM_MAX_SENS_SITES; i++ )
            sens |= (site == AMIRD_SENSITIVE_SITES[i]) ? TRUE : FALSE;

        // Test whether the same group type
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = sens;
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "grouptype") == 0 )
    {
        // Check argumements, get test group type
        AString gttst;
        if ( (APolicyImplementor::NoArgs(cond) != 1) || 
             (APolicyImplementor::Arg(0,cond,gttst) == FALSE) )
        {
            // Error out
            LOGMSG1( log, AGRP_ALEVEL, "Malformed groupType "
                     "condition [%s]", (char *)cond );
            return( FALSE );
        }

        // Get the group class (if needed)
        if ( (groupClass.Length() == 0) &&
             (Environment::getStringEnvVar("GROUP", groupClass) == FALSE) )
            groupClass = "unknown";

        // Test whether the same group type
        cond.ResultType() = ACondEval::APDL_INVARIANT;
        cond.Result() = (gttst == groupClass);
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate INVARIENT %s", (char *)pred, 
                 BOOLE_STR(cond.Result()) );
    }
    else if ( strcasecmp((char *)pred, "notlockdown") == 0 )
    {
        // OK check to see if system locked down
        struct stat finfo;
        cond.ResultType() = ACondEval::APDL_ALWAYS_EVAL;
        cond.Result() = (stat("./lock",&finfo) == 0) ? FALSE : TRUE;
        LOGMSG2( log, AGRP_ALEVEL, "%s Evaluate ALWAYS_EVAL %s", 
                 (char *)cond, BOOLE_STR(cond.Result()) );
    }
    else
    {
        // Unknown predicate, not implemented
        return( FALSE );
    }

    // Return evaluated
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : AMirDPolicyImpl::testPredicates
// Description  : test some of the implentor predicates
//
// Inputs       : none
// Outputs      : none

void AMirDPolicyImpl::testPredicates( void )
{
    // Setup some test predicates
    const char * cevals[3] = { 
        "iscontrolgroup()",
        "isdownloadgroup()",
        "hassensitivefilesystem()" };
    const char * devals[12] = { 
        "iscontrolgroup()",
        "isdownloadgroup()",
        "hasReadAccess(root,/etc/passwd)",
        "hasReadAccess(pdmcdan,/usr/local/src/antigone-2.0)",
        "hasReadAccess(pdmcdan,/root/notes)",
        "hasReadAccess(pdmcdan,/home/pdmcdan/xroierlsdkj)",
        "hasReadAccess(one,/home/pdmcdan/nsmail)",
        "hasReadAccess(one,/home/pdmcdan/admin)",
        "issensitivefile(/etc/passwd)",
        "issensitivefile(/home/pdmcdan/.profile)",
        "issensitivefile(/home/pdmcdan/thesis/bibliography)",
        "injoinphase()" };

    // If this is a control group
    const char ** nvls = (groupType == AMIRD_CONTROL_GROUP) ? cevals : devals;
    int max = (groupType == AMIRD_CONTROL_GROUP) ? 3 : 12;
    APolicyInst *inst = (groupType == AMIRD_CONTROL_GROUP) ?
        cntrlGroup->ControlGroup()->PolicyInst() :
        dwnldGroup->Group()->PolicyInst();

    // Setup the evals
    for ( int i=0; i<max; i++ )
    {
        // Setup, do, then log the evaluation
        AString cond( nvls[i] );
        ACondEval ev( *inst, cond );
        Boolean res = EvaluateCondition( ev );
        LOGMSG2( log, AGRP_ALEVEL, "Eval %s [success = %s]", 
                 nvls[i], BOOLE_STR(res) );
    }

    // Return, no return code
    return;
}
