/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : amird.cpp
//
//   Created : 25-Sep-00
//   By : Patrick McDaniel
//
//   Description : This is the main code file for the amird Secure
//                 Mirroring application.  See the associated
//                 documentation for details on its use.
//
//   *** Change History ***
//
//   25-Sep-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <signal.h>
#include <Environment.h>

// Application Include Files
#include <AMirD.h>

// Defines
#define USAGE \
"USAGE % amird [-c=<config>] [-l=log] [-v=<log level>]\n" \
"where \n" \
"    <config>    - amird configuration file\n" \
"    <log>       - amird log file\n\n" \
"    <log level> - amird log level\n\n"

//
// Global Data
AMirD *amird = NULL; // The mirroring application

//
// Functions

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : sigHandler
// Description  : signal handler
//
// Inputs       : sig - signal to handle
// Outputs      : none

void sigHandler( int sig )
{
    cerr << "Got signal (" << sig << ")" << endl;
    cout << endl << flush;

    // Switch on signal type
    switch( sig )
    {
        case SIGHUP: // Re-read the configuration
             amird->reconfigure();
             break;

        case SIGUSR1: // Start refresh
             amird->resync();
             break;

        case SIGINT:  // Interrupt
        case SIGQUIT: // Quititing application
             amird->shutDown(); // Shutdown and exit
             delete amird;
             exit( 0 );
    }

    // On Linux, we need to reset the handler each time it is used
    signal( sig, sigHandler );

    // Return, no return code
    return;
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : main
// Description  : main function
//
// Inputs       : argc - input arguments
//                argv - pointers to input arguments
// Outputs      : 0 if successful, -1 if failure

int main( int argc, char **argv )
{
    // Create some definitions for CLP and ENV
    char *clps[] = { "c=s", "l=s", "v=i" };
    char *usage = USAGE; 

    // Setup signal handlers, execute
    if ( (signal( SIGHUP, sigHandler ) == SIG_ERR) ||
         (signal( SIGUSR1, sigHandler ) == SIG_ERR) ||
         (signal( SIGINT, sigHandler ) == SIG_ERR) ||
         (signal( SIGQUIT, sigHandler ) == SIG_ERR) )
    {
        // Log error
        cerr << "Unable to set signal handlers" << endl;
        cerr << "Errpor : " << strerror(errno) << endl;
    }

    // Always check for exception in processing
    Boolean retcode;
    Environment *env;
    AString config, logname;
    try
    {
        // Create the environment, get arguments
        env = new Environment( 3, clps, 0, NULL, argc, argv, usage );

        // Get configuration filename (or accept default)
        if ( env->clpHasValue("c") )
            config = env->getStringCLPValue("c");
        else
            config = AMIRD_DEFAULT_CNF;

        // Get log filename (or accept default)
        if ( env->clpHasValue("l") )
            logname = env->getStringCLPValue("l");
        else
            logname = AMIRD_DEFAULT_LOG;

        // Create the amird application
        amird = new AMirD( config, &logname );

        // Set the logging level
        if ( amird->Log() != NULL )
            amird->Log()->setAuditLevel( (env->clpHasValue("v")) ? 
                                          env->getIntCLPValue("v") : 
                                         AMIRD_ALEVEL|FXFER_REPORT_LEVEL|FATAL_ERROR|MAJOR_ERROR/*|APDL_ALEVEL|APDL_DETAIL|AMECH_ALEVEL*/ );
        amird->Log()->setShowLogLevel(TRUE);    // ++++
        amird->Log()->setShowSource(TRUE);    // ++++
        amird->Log()->setAuditLevel(-1 & (~ATRANS_DETAIL) & (~APDL_DETAIL) & (~AMECH_ALEVEL) & (~AMECH_DETAIL));    // ++++
        amird->Log()->setTimestamping(FALSE);    // ++++

        // Execute the event loop
        retcode = amird->execute();
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        cerr << (char *)e->toString() << ", aborting." << endl;
        cerr << usage << endl;
        return( -1 );
    }

    // Exit success/failure return code
    return( (retcode == TRUE) ? 0 : -1 );
}
