/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : xfer.cpp
//
//   Created : 09-Oct-00
//   By : Patrick McDaniel
//
//   Description : This is the demonstation application for the file
//                 transfer classes.
//
//   *** Change History ***
//
//   09-Oct-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files
#include <errno.h>
#include <Environment.h>

// Project Include Files
#include <AMFxfer.h>

//
// Global Data

// The mirroring application.  Not used for this test, just makes the code link.
void *amird = NULL;

// Defines
#define USAGE \
"USAGE xfer <mc addr> <port> <loc policy> <fsys> <file> [-p group policy] " \
"[-l=<log>] -o -t\n" \
"\n" \
"where " \
"\n" \
"    mc address    - multicast address \n" \
"    port          - port associated with mc/srv address\n" \
"    loc policy    - file containing local policy\n" \
"    file          - file to xfer/recv\n" \
"    group policy  - file containing group policy (server only)\n" \
"    log           - filename for application log\n" \
"    t             - timestamp logs\n" \
"    o             - enable mcast loopback\n\n"

#define PE_DEFAULT_ALEVEL DEFAULT_ALEVEL|AMIRD_ALEVEL
#define DEF_LOGNAME "xfer.log"

// Static Application Data
AuditLog *alog = NULL;

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : executeXfer
// Description  : Execute the file transfer
//
// Inputs       : xfer - file transfer object
// Outputs      : TRUE if successful, FALSE otherwise

Boolean executeXfer( AMFxfer * xfer )
{
    //
    // Local variables
    struct timeval tm;
    fd_set rfds;
    int n, retval;

    // While still something to do, keep processing
    xfer->doXfer();
    while ( xfer->Status() == AMFxfer::OK )
    {
        // Clear select vars
        n = 0;
        FD_ZERO( &rfds );
        tm.tv_sec = MAX_AMIRD_WAIT;
        tm.tv_usec = 0;

        // If no messages pending
        if ( xfer->messagePending() == FALSE )
        {
            // Setup a select call, check result
            xfer->fdSet( n, &rfds, NULL, NULL, &tm );
LOGMSG2( alog, FATAL_ERROR, "In select (%ld,%ld)", tm.tv_sec, tm.tv_usec );
            retval = select( n, &rfds, NULL, NULL, &tm );
LOGMSG1( alog, FATAL_ERROR, "out (%d)", retval );
            if ( retval < 0 )
            {
                // Report error and return
                LOGMSG2( alog, FATAL_ERROR, "Bad select call (%d)[%s]", 
                         errno, strerror(errno) )
                return( FALSE );
            }
        }

        // Do the xfer stuff
        xfer->doXfer( &rfds );
    }

    // Return successfully (or not)
    return( xfer->Status() == AMFxfer::XFERCOMPL );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : main
// Description  : main function
//
// Inputs       : argc - input arguments
//                argv - pointers to input arguments
// Outputs      : 0 if successful, -1 if failure

int main( int argc, char **argv )
{
    // Create some definitions for CLP and ENV
    char * clps[] = { "!<1>=s", "!<2>=i", "!<3>=s", "!<4>=s", 
                      "p=s", "l=s", "t=b", "o=b" };
    char *usage = USAGE; 

    // Always check for exception in processing
    Environment *env;
    AString alogs = DEF_LOGNAME, gaddr, gpol, lpol, fn, rl;
    int gport;
    ATransport *trans;
    AMFxfer * xfer;
    Boolean isServer = FALSE;
    try
    {
        // Create the environment, get arguments
        env = new Environment( 8, clps, 0, NULL, argc, argv, usage );

        // Set timestamping on/off
        Boolean ts = env->clpHasValue("t");

        // Create log (with or without file echo)
        if ( env->clpHasValue("l") )
        {
            // Create log with file echo
            alogs = env->getStringCLPValue("l");
            alog = new AuditLog( alogs, TRUE, TRUE, ts );
        }
        else
            // Create log without file echo
            alog = new AuditLog( NULL, TRUE, FALSE, ts );
        alog->setAuditLevel( PE_DEFAULT_ALEVEL );
        AMEntry::setEntryLog( alog );

        // Get the group address (or server address)
        gaddr = env->getStringCLPValue("<1>");
        gport = env->getIntCLPValue("<2>");
        lpol = env->getStringCLPValue("<3>");
        fn = env->getStringCLPValue("<4>");
        isServer = env->clpHasValue("p");

        // First, setup the transport layer address
        IPAddress *gipaddr = IPAddress::IPAddressFactory( gaddr, gport );
        if ( gipaddr == NULL )
        {
            // Error out
            LOGMSG1( alog, FATAL_ERROR, "Unable to interpret group address [%s]", 
                     (char *)gaddr );
            return( -1 );
        }

        // Now create the transport, then group
        if ( isServer == TRUE )
        {
            // Server, create transfer then xfer
            trans = new ATransport( gipaddr, (unsigned short)0, 
                                    ATransport::AT_SYMMETRIC );
            gpol = env->getStringCLPValue("p");
            xfer = new AMFxfer( trans, fn, gpol, lpol, alog );
        }
        else
        {
            // Client, use group or asym addresss, then create group
            trans = new ATransport( gipaddr, (IPAddress *)NULL, 
                                    ATransport::AT_SYMMETRIC );
            xfer = new AMFxfer( trans, fn, lpol, alog );
        }

        // Set the multicast loopback policy (if specified)
        if ( env->clpHasValue("o") ) trans->SetLoopback( TRUE );

        // Check the status of the xfer object
        if ( xfer->Status() != AMFxfer::OK )
        {
            // Log and error out
            LOGMSG1( alog, FATAL_ERROR, "Bad state on xfer [%s]", 
                     xfer->statusString(xfer->Status()) );
            return( -1 );
        }

        // Execute the file transfer
        executeXfer( xfer );
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        LOGMSG1( alog, FATAL_ERROR, "Exception [%s]", e->Description() );
        cerr << (char *)e->toString() << ", aborting." << endl;
        cerr << usage << endl;
        return( -1 );
    }

    // Cleanup
    delete env;
    delete xfer;
    delete trans;
    delete alog;

    // Exit success/failure return code
    return( 0 );
}
